from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.action import SendMessageAction
from aasm.intermediate.behaviour import MessageReceivedBehaviour
from aasm.utils.validation import is_connection, is_float, is_int

if TYPE_CHECKING:
    from typing import List as TypingList
    from typing import Type

    from parsing.state import State


class ArgumentType:
    ...


class Mutable(ArgumentType):
    ...


class Immutable(ArgumentType):
    ...


class Declared(ArgumentType):
    ...


class Float(ArgumentType):
    ...


class Integer(ArgumentType):
    ...


class Enum(ArgumentType):
    ...


class EnumValue(ArgumentType):
    ...


class List(ArgumentType):
    ...


class ConnectionList(ArgumentType):
    ...


class MessageList(ArgumentType):
    ...


class FloatList(ArgumentType):
    ...


class AgentParam(ArgumentType):
    ...


class Message(ArgumentType):
    ...


class MessageType(ArgumentType):
    ...


class ReceivedMessage(ArgumentType):
    ...


class ReceivedMessageParam(ArgumentType):
    ...


class SendMessage(ArgumentType):
    ...


class SendMessageParam(ArgumentType):
    ...


class Connection(ArgumentType):
    ...


class Literal(ArgumentType):
    ...


class Argument:
    """Doesn't panic. Use in the action context."""

    def __init__(self, state: State, expr: str):
        self.expr: str = expr
        self.types: TypingList[ArgumentType] = []
        self.type_in_op: ArgumentType | None = None
        self.set_types(state)

    def set_types(self, state: State) -> None:
        self.check_agent_params(state)
        self.check_action_variables(state)
        self.check_received_message_params(state)
        self.check_send_message_params(state)
        self.check_numerical_values()
        self.check_connection_values()

    def check_agent_params(self, state: State) -> None:
        if self.expr in state.last_agent.RESERVED_FLOAT_PARAMS:
            self.types.append(self.compose(Float, AgentParam, Immutable))

        elif self.expr in state.last_agent.RESERVED_CONNECTION_LIST_PARAMS:
            self.types.append(self.compose(List, ConnectionList, AgentParam, Mutable))

        elif self.expr in state.last_agent.float_param_names:
            self.types.append(self.compose(Float, AgentParam, Mutable))

        elif self.expr in state.last_agent.enums:
            self.types.append(self.compose(Enum, AgentParam, Mutable))

        elif self.expr in state.last_agent.message_lists:
            self.types.append(self.compose(List, MessageList, AgentParam, Mutable))

        elif self.expr in state.last_agent.connection_lists:
            self.types.append(self.compose(List, ConnectionList, AgentParam, Mutable))

        elif self.expr in state.last_agent.float_lists:
            self.types.append(self.compose(List, FloatList, AgentParam, Mutable))

        for enum_param in state.last_agent.enums.values():
            for enum_value in enum_param.enum_values:
                if self.expr == enum_value.value:
                    self.types.append(
                        self.compose(EnumValue, Immutable, from_enum=enum_param.name)
                    )

    def check_action_variables(self, state: State) -> None:
        if state.last_action.is_declared_float(self.expr):
            self.types.append(self.compose(Float, Declared, Mutable))

        elif state.last_action.is_declared_connection(self.expr):
            self.types.append(self.compose(Connection, Declared, Mutable))

    def check_numerical_values(self) -> None:
        if is_float(self.expr):
            self.types.append(self.compose(Float, Immutable, Literal))

        if is_int(self.expr):
            self.types.append(self.compose(Integer, Immutable, Literal))

    def check_connection_values(self) -> None:
        if is_connection(self.expr):
            self.types.append(self.compose(Connection, Immutable, Literal))

    def check_received_message_params(self, state: State) -> None:
        if isinstance(state.last_behaviour, MessageReceivedBehaviour):
            if self.expr.lower().startswith("rcv."):
                prop = self.expr.split(".")[1]

                if (
                    prop
                    in state.last_behaviour.received_message.RESERVED_CONNECTION_PARAMS
                ):
                    self.types.append(
                        self.compose(Connection, ReceivedMessageParam, Immutable)
                    )

                elif prop in state.last_behaviour.received_message.RESERVED_TYPE_PARAMS:
                    self.types.append(
                        self.compose(MessageType, ReceivedMessageParam, Immutable)
                    )

                elif prop in state.last_behaviour.received_message.float_params:
                    self.types.append(
                        self.compose(Float, ReceivedMessageParam, Immutable)
                    )

                elif prop in state.last_behaviour.received_message.connection_params:
                    self.types.append(
                        self.compose(Connection, ReceivedMessageParam, Immutable)
                    )

            elif self.expr.lower() == "rcv":
                self.types.append(self.compose(Message, ReceivedMessage, Immutable))

    def check_send_message_params(self, state: State) -> None:
        if isinstance(state.last_action, SendMessageAction):
            if self.expr.lower().startswith("send."):
                prop = self.expr.split(".")[1]

                if prop in state.last_action.send_message.RESERVED_TYPE_PARAMS:
                    self.types.append(
                        self.compose(MessageType, SendMessageParam, Immutable)
                    )

                elif prop in state.last_action.send_message.float_params:
                    self.types.append(self.compose(Float, SendMessageParam, Mutable))

                elif prop in state.last_action.send_message.connection_params:
                    self.types.append(
                        self.compose(Connection, SendMessageParam, Mutable)
                    )

            elif self.expr.lower() == "send":
                self.types.append(self.compose(Message, SendMessage, Mutable))

    def compose(self, *classes: Type[ArgumentType], **args: str) -> ArgumentType:
        name = "_".join([klass.__name__ for klass in classes])
        if args:
            name += "|" + "|".join([f"{key}={value}" for key, value in args.items()])
        return type(name, classes, args)()

    def has_arg(self, argument_type: ArgumentType, key: str, value: str) -> bool:
        try:
            if getattr(argument_type, key) == value:
                return True
        except AttributeError:
            ...
        return False

    def has_type(self, *classes: Type[ArgumentType], **args: str) -> bool:
        for type_ in self.types:
            if all([isinstance(type_, klass) for klass in classes]) and all(
                [self.has_arg(type_, key, value) for key, value in args.items()]
            ):
                return True
        return False

    def set_op_type(self, *classes: Type[ArgumentType], **args: str) -> None:
        for type_ in self.types:
            if all([isinstance(type_, klass) for klass in classes]) and all(
                [self.has_arg(type_, key, value) for key, value in args.items()]
            ):
                self.type_in_op = type_

    # DECL
    def declaration_context(self, rhs: Argument) -> bool:
        if rhs.has_type(Float):
            self.type_in_op = self.compose(Float, Declared, Mutable)
            rhs.set_op_type(Float)

        elif rhs.has_type(Connection):
            self.type_in_op = self.compose(Connection, Declared, Mutable)
            rhs.set_op_type(Connection)

        else:
            return False

        return True

    # IEQ, INEQ, WEQ, WNEQ
    def unordered_comparaison_context(self, rhs: Argument) -> bool:
        if self.has_type(Float) and rhs.has_type(Float):
            self.set_op_type(Float)
            rhs.set_op_type(Float)

        elif self.has_type(Enum) and rhs.has_type(EnumValue, from_enum=self.expr):
            self.set_op_type(Enum)
            rhs.set_op_type(EnumValue, from_enum=self.expr)

        else:
            return False

        return True

    # IGT, IGTEQ, ILT, ILTEQ, WGT, WGTEQ, WLT, WLTEQ
    def ordered_comparaison_context(self, rhs: Argument) -> bool:
        if self.has_type(Float) and rhs.has_type(Float):
            self.set_op_type(Float)
            rhs.set_op_type(Float)
            return True

        return False

    # ADD, SUBT, MULT, DIV, SIN, COS
    def math_context(self, rhs: Argument) -> bool:
        if self.has_type(Float, Mutable) and rhs.has_type(Float):
            self.set_op_type(Float, Mutable)
            rhs.set_op_type(Float)
            return True

        return False

    # POW, LOG
    def math_exponentiation_context(self, base: Argument, arg: Argument) -> bool:
        if (
            self.has_type(Float, Mutable)
            and base.has_type(Float)
            and arg.has_type(Float)
        ):
            self.set_op_type(Float, Mutable)
            base.set_op_type(Float)
            arg.set_op_type(Float)
            return True

        return False

    # MOD
    def math_modulo_context(self, dividend: Argument, divisor: Argument) -> bool:
        if (
            self.has_type(Float, Mutable)
            and dividend.has_type(Float)
            and divisor.has_type(Float)
        ):
            self.set_op_type(Float, Mutable)
            dividend.set_op_type(Float)
            divisor.set_op_type(Float)
            return True

        return False

    # ADDE, REME
    def list_modification_context(self, rhs: Argument) -> bool:
        if self.has_type(ConnectionList, Mutable) and rhs.has_type(Connection):
            self.set_op_type(ConnectionList, Mutable)
            rhs.set_op_type(Connection)

        elif self.has_type(MessageList, Mutable) and rhs.has_type(Message):
            self.set_op_type(MessageList, Mutable)
            rhs.set_op_type(Message)

        elif self.has_type(FloatList, Mutable) and rhs.has_type(Float):
            self.set_op_type(FloatList, Mutable)
            rhs.set_op_type(Float)

        else:
            return False

        return True

    # REMEN
    def list_n_removal_context(self, rhs: Argument) -> bool:
        if self.has_type(List, Mutable) and rhs.has_type(Float):
            self.set_op_type(List, Mutable)
            rhs.set_op_type(Float)
            return True

        return False

    # SET
    def assignment_context(self, rhs: Argument) -> bool:
        if self.has_type(Enum, Mutable) and rhs.has_type(
            EnumValue, from_enum=self.expr
        ):
            self.set_op_type(Enum, Mutable)
            rhs.set_op_type(EnumValue, from_enum=self.expr)

        elif self.has_type(Float, Mutable) and rhs.has_type(Float):
            self.set_op_type(Float, Mutable)
            rhs.set_op_type(Float)

        elif self.has_type(Message, Mutable) and rhs.has_type(MessageList):
            self.set_op_type(Message, Mutable)
            rhs.set_op_type(MessageList)

        elif self.has_type(SendMessageParam, Mutable) and rhs.has_type(Float):
            self.set_op_type(SendMessageParam, Mutable)
            rhs.set_op_type(Float)

        elif self.has_type(Connection, Mutable) and rhs.has_type(Connection):
            self.set_op_type(Connection, Mutable)
            rhs.set_op_type(Connection)

        else:
            return False

        return True

    # SUBS
    def list_subset_context(self, from_list: Argument, num: Argument) -> bool:
        if (
            self.has_type(ConnectionList, Mutable)
            and from_list.has_type(ConnectionList)
            and num.has_type(Float)
        ):
            self.set_op_type(ConnectionList, Mutable)
            from_list.set_op_type(ConnectionList)
            num.set_op_type(Float)
            return True

        return False

    # IN, NIN
    def list_inclusion_context(self, rhs: Argument) -> bool:
        if self.has_type(ConnectionList) and rhs.has_type(Connection):
            self.set_op_type(ConnectionList)
            rhs.set_op_type(Connection)

        elif self.has_type(MessageList) and rhs.has_type(Message):
            self.set_op_type(MessageList)
            rhs.set_op_type(Message)

        elif self.has_type(FloatList) and rhs.has_type(Float):
            self.set_op_type(FloatList)
            rhs.set_op_type(Float)

        else:
            return False

        return True

    # CLR
    def list_clear_context(self) -> bool:
        if self.has_type(List, Mutable):
            self.set_op_type(List, Mutable)
            return True

        return False

    # LEN
    def list_length_context(self, rhs: Argument) -> bool:
        if self.has_type(Float, Mutable) and rhs.has_type(List):
            self.set_op_type(Float, Mutable)
            rhs.set_op_type(List)
            return True

        return False

    # SEND
    def send_context(self) -> bool:
        if self.has_type(ConnectionList):
            self.set_op_type(ConnectionList)

        elif self.has_type(Connection):
            self.set_op_type(Connection)

        else:
            return False

        return True

    # RAND
    def random_number_generation_context(self, *args: Argument) -> bool:
        if self.has_type(Float, Mutable) and all([arg.has_type(Float) for arg in args]):
            self.set_op_type(Float, Mutable)
            for arg in args:
                arg.set_op_type(Float)
            return True

        return False

    # ROUND
    def round_number_context(self) -> bool:
        if self.has_type(Float, Mutable):
            self.set_op_type(Float, Mutable)
            return True

        return False

    # LR
    def list_read_context(self, src: Argument, idx: Argument) -> bool:
        if (
            self.has_type(Float, Mutable)
            and src.has_type(FloatList)
            and idx.has_type(Float)
        ):
            self.set_op_type(Float, Mutable)
            src.set_op_type(FloatList)
            idx.set_op_type(Float)

        elif (
            self.has_type(Connection, Mutable)
            and src.has_type(ConnectionList)
            and idx.has_type(Float)
        ):
            self.set_op_type(Connection, Mutable)
            src.set_op_type(ConnectionList)
            idx.set_op_type(Float)

        else:
            return False

        return True

    # LW
    def list_write_context(self, idx: Argument, value: Argument) -> bool:
        if (
            self.has_type(FloatList, Mutable)
            and idx.has_type(Float)
            and value.has_type(Float)
        ):
            self.set_op_type(FloatList, Mutable)
            idx.set_op_type(Float)
            value.set_op_type(Float)

        elif (
            self.has_type(ConnectionList, Mutable)
            and idx.has_type(Float)
            and value.has_type(Connection)
        ):
            self.set_op_type(ConnectionList, Mutable)
            idx.set_op_type(Float)
            value.set_op_type(Connection)

        else:
            return False

        return True

    def explain(self) -> str:
        types = f"{self.expr}: [ "
        for argument_type in self.types:
            types += type(argument_type).__name__ + ", "
        types = types.rstrip().rsplit(",", 1)[0]
        types += " ]"
        return types

    def print(self) -> None:
        print(f"Argument {self.expr}")
        print(
            f"Type in op: {type(self.type_in_op).__name__ if self.type_in_op else 'UNKNOWN'}"
        )
        for argument_type in self.types:
            type(argument_type).__name__



from typing import NoReturn


class PreprocessorItem:
    def __init__(self, signature: str):
        self.signature = signature

    def expand(self) -> NoReturn:
        raise NotImplementedError

    def add_definition(self, definition: str) -> NoReturn:
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.agent import Agent
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_AGENT(state: State, name: str) -> None:
    state.require(
        not state.in_agent,
        "Already inside an agent.",
        "First end current agent using EAGENT.",
    )
    state.require(
        not state.in_message,
        "Cannot define agents inside messages.",
        "First end current message using EMESSAGE.",
    )
    state.require(
        not state.in_graph,
        "Cannot define agents inside graphs.",
        "First end current graph using EGRAPH.",
    )
    state.require(
        not state.agent_exists(name),
        f"Agent {name} already exists in the current environment.",
    )
    state.require(
        is_valid_name(name),
        f"{name} is not a correct name.",
        f"Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.",
    )

    state.in_agent = True
    state.add_agent(Agent(name))


def op_EAGENT(state: State) -> None:
    state.require(
        state.in_agent, "Not inside any agent.", "Try defining new agents using AGENT."
    )
    state.require(
        not state.in_behaviour,
        "Cannot end an agent inside a behaviour.",
        "First end current behaviour using EBEHAV.",
    )

    state.in_agent = False



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aasm.intermediate.argument import Argument


class Declaration:
    def __init__(self, name: Argument, value: Argument):
        self.name: str = name.expr
        self.value: Argument = value

    def print(self) -> None:
        print(f"Declaration: {self.name}")
        self.value.print()


class FloatDeclaration(Declaration):
    def __init__(self, name: Argument, value: Argument):
        super().__init__(name=name, value=value)

    def print(self) -> None:
        print(f"FloatDeclaration: {self.name}")
        super().print()


class ConnectionDeclaration(Declaration):
    def __init__(self, name: Argument, value: Argument):
        super().__init__(name=name, value=value)

    def print(self) -> None:
        print(f"ConnectionDeclaration: {self.name}")
        super().print()



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Round

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_ROUND(state: State, arg1: str) -> None:
    state.require(
        state.in_action, "Not inside any action.", f"ROUND can be used inside actions."
    )
    num = Argument(state, arg1)
    state.require(
        num.round_number_context(),
        "Mismatched type in the round number context.",
        f"NUM {num.explain()}",
    )

    state.last_action.add_instruction(Round(num))



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import ExpDist, NormalDist, UniformDist
from aasm.parsing.op.round import op_ROUND

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_RAND(state: State, arg1: str, arg2: str, arg3: str, args: List[str]) -> None:
    state.require(
        state.in_action, "Not inside any action.", f"RAND can be used inside actions."
    )
    result = Argument(state, arg1)

    match arg3, args:
        case "uniform", [a, b]:
            a_arg = Argument(state, a)
            b_arg = Argument(state, b)
            state.require(
                result.random_number_generation_context(a_arg, b_arg),
                "Mismatched types in the random number generation context.",
                f"RESULT {result.explain()}, A {a_arg.explain()}, B {b_arg.explain()}",
            )

            state.last_action.add_instruction(UniformDist(result, a_arg, b_arg))

        case "normal", [mean, std_dev]:
            mean_arg = Argument(state, mean)
            std_dev_arg = Argument(state, std_dev)
            state.require(
                result.random_number_generation_context(mean_arg, std_dev_arg),
                "Mismatched types in the random number generation context.",
                f"RESULT {result.explain()}, MEAN {mean_arg.explain()}, STD_DEV {std_dev_arg.explain()}",
            )

            state.last_action.add_instruction(NormalDist(result, mean_arg, std_dev_arg))

        case "exp", [lambda_]:
            lambda_arg = Argument(state, lambda_)
            state.require(
                result.random_number_generation_context(lambda_arg),
                "Mismatched types in the random number generation context.",
                f"RESULT {result.explain()}, LAMBDA {lambda_arg.explain()}",
            )

            state.last_action.add_instruction(ExpDist(result, lambda_arg))

        case _:
            state.panic(f"Incorrect operation: RAND {arg1} {arg2} {arg3} {args}")

    match arg2:
        case "float":
            ...

        case "int":
            op_ROUND(state, arg1)

        case _:
            state.panic(f"Incorrect operation: RAND {arg1} {arg2} {arg3} {args}")



import keyword
from decimal import Decimal
from typing import List

from aasm.utils.iteration import zip_consecutive_pairs


def is_float(value: str) -> bool:
    try:
        float(value)
        return True
    except ValueError:
        return False


def is_int(value: str) -> bool:
    return value.isdigit() or (value[0] == "-" and value[1:].isdigit())


def is_connection(value: str) -> bool:
    return isinstance(value, str) and value.startswith('"') and value.endswith('"')


def is_valid_enum_list(enums: List[str]) -> bool:
    if len(enums) == 0 or len(enums) % 2:
        return False
    total_sum = Decimal(0.0)
    for enum_pair in zip_consecutive_pairs(enums):
        if not is_float(enum_pair[1]) or float(enum_pair[1]) < 0.0:
            return False
        total_sum += Decimal(float(enum_pair[1]))
    if total_sum < 99.0 or total_sum > 101.0:
        return False
    return True


def is_valid_name(name: str) -> bool:
    return (
        len(name) != 0
        and not name[0].isdigit()
        and (name.isalnum() or "_" in name)
        and name.lower() not in get_invalid_names()
    )


def get_invalid_names() -> List[str]:
    invalid_names = [
        "send",
        "rcv",
        "len",
        "round",
        "list",
        "filter",
        "self",
        "jid",
        "datetime",
        "random",
        "numpy",
        "json",
        "spade",
        "copy",
        "uuid",
        "get_json_from_spade_message",
        "get_spade_message",
        "logger",
        "any",
        "sys",
        "limit_number",
        "int",
        "BackupBehaviour",
        "backup_url",
        "backup_period",
        "backup_delay",
        "setup",
    ]
    invalid_names.extend(keyword.kwlist)
    return invalid_names


def print_invalid_names() -> str:
    return ", ".join(get_invalid_names())



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.graph import MatrixGraph, StatisticalGraph

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_GRAPH(state: State, category: str) -> None:
    state.require(
        not state.in_graph,
        "Already inside a graph.",
        "First end current graph using EGRAPH.",
    )
    state.require(
        not state.in_agent,
        "Cannot define graphs inside agents.",
        "First end current message using EAGENT.",
    )
    state.require(
        not state.in_message,
        "Cannot define graphs inside messages.",
        "First end current message using EMESSAGE.",
    )
    state.require(
        not state.graph_exists(), "Graph already exists in the current environment."
    )

    match category:
        case "statistical":
            state.add_graph(StatisticalGraph())
        case "matrix":
            state.add_graph(MatrixGraph())
        case _:
            state.panic(f"Incorrect operation: GRAPH {category}")

    state.in_graph = True


def op_EGRAPH(state: State) -> None:
    state.require(
        state.in_graph, "Not inside any graph.", "Try defining new graphs using GRAPH."
    )
    if (
        isinstance(state.last_graph, StatisticalGraph)
        and state.last_graph.is_agent_percent_amount_used()
    ):
        state.require(
            state.last_graph.is_size_defined(),
            "Graph size is not defined.",
            "Graph size must be defined to use agent percent amount.",
        )

    state.in_graph = False



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.generating.code import Code
from aasm.generating.python_code import PythonCode
from aasm.generating.python_graph import PythonGraph
from aasm.intermediate.action import SendMessageAction
from aasm.intermediate.argument import (
    AgentParam,
    Connection,
    ConnectionList,
    EnumValue,
    Float,
    FloatList,
    MessageList,
    ReceivedMessageParam,
    SendMessageParam,
)
from aasm.intermediate.behaviour import MessageReceivedBehaviour
from aasm.intermediate.block import Block
from aasm.intermediate.declaration import ConnectionDeclaration, FloatDeclaration
from aasm.intermediate.instruction import (
    Add,
    AddElement,
    Clear,
    Comparaison,
    Cos,
    Divide,
    ExpDist,
    ExponentiationOperation,
    IfEqual,
    IfGreaterThan,
    IfGreaterThanOrEqual,
    IfInList,
    IfLessThan,
    IfLessThanOrEqual,
    IfNotEqual,
    IfNotInList,
    Length,
    ListElementAccess,
    ListRead,
    ListWrite,
    Logarithm,
    MathOperation,
    Modulo,
    Multiply,
    NormalDist,
    Power,
    RemoveElement,
    RemoveNElements,
    Round,
    Send,
    Set,
    Sin,
    Subset,
    Subtract,
    TrigonometryOperation,
    UniformDist,
    WhileEqual,
    WhileGreaterThan,
    WhileGreaterThanOrEqual,
    WhileLessThan,
    WhileLessThanOrEqual,
    WhileNotEqual,
)
from aasm.parsing.parse import parse_lines

if TYPE_CHECKING:
    from aasm.intermediate.action import Action
    from aasm.intermediate.agent import Agent
    from aasm.intermediate.argument import Argument
    from aasm.intermediate.behaviour import Behaviour
    from aasm.intermediate.message import Message as IntermediateMessage


def get_spade_code(
    aasm_lines: List[str], indent_size: int = 4, debug: bool = False
) -> Code:
    """Generates SPADE code in Python from `aasm_lines`.

    Parameters
    ----------
    aasm_lines: List[str]
        Lines of code written in Agents Assembly

    indent_size: int, optional
        Python code indentation size

    debug: bool, optional
        Print the translator debug information to the standard output

    Returns
    -------
    SPADE code along with the algorithm for the graph generation

    Raises
    ------
    PanicException
        If an error is detected while parsing the `aasm_lines`.
    """
    parsed = parse_lines(aasm_lines, debug)
    return Code(
        PythonSpadeCode(indent_size, parsed.agents).code_lines,
        PythonGraph(indent_size, parsed.graph).code_lines,
    )


class PythonSpadeCode(PythonCode):
    def __init__(self, indent_size: int, agents: List[Agent]):
        super().__init__(indent_size)
        if agents:
            self.add_required_imports()
            for agent in agents:
                self.add_newlines(2)
                self.generate_agent(agent)

    def add_required_imports(self) -> None:
        self.add_line("import copy")
        self.add_line("import datetime")
        self.add_line("import random")
        self.add_line("import httpx")
        self.add_line("import numpy")
        self.add_line("import orjson")
        self.add_line("import spade")
        self.add_line("import sys")

    def generate_agent(self, agent: Agent) -> None:
        self.add_line(f"class {agent.name}(spade.agent.Agent):")
        self.indent_right()

        self.add_agent_constructor(agent)
        self.add_newline()

        self.add_float_utils()
        self.add_newline()

        self.add_message_utils()
        self.add_newline()

        self.add_agent_setup(agent)
        self.add_newline()

        self.add_backup_behaviour(agent)
        self.add_newline()

        for setup_behaviour in agent.setup_behaviours.values():
            self.add_agent_behaviour(
                setup_behaviour, "spade.behaviour.OneShotBehaviour"
            )
            self.add_newline()

        for one_time_behaviour in agent.one_time_behaviours.values():
            self.add_agent_behaviour(
                one_time_behaviour, "spade.behaviour.TimeoutBehaviour"
            )
            self.add_newline()

        for cyclic_behaviour in agent.cyclic_behaviours.values():
            self.add_agent_behaviour(
                cyclic_behaviour, "spade.behaviour.PeriodicBehaviour"
            )
            self.add_newline()

        for message_received_behaviour in agent.message_received_behaviours.values():
            self.add_agent_behaviour(
                message_received_behaviour, "spade.behaviour.CyclicBehaviour"
            )
            self.add_newline()

        self.indent_left()

    def add_agent_constructor(self, agent: Agent) -> None:
        self.add_line(
            "def __init__(self, jid, password, backup_method = None, backup_queue = None, backup_url = None, backup_period = 60, backup_delay = 0, logger = None, **kwargs):"
        )
        self.indent_right()

        self.add_line("super().__init__(jid, password, verify_security=False)")
        self.add_line(
            "if logger: logger.debug(f'[{jid}] Received parameters: jid: {jid}, password: {password}, backup_method: {backup_method}, backup_queue: {backup_queue}, backup_url: {backup_url}, backup_period: {backup_period}, backup_delay: {backup_delay}, kwargs: {kwargs}')"
        )
        self.add_line("self.logger = logger")
        self.add_line("self.backup_method = backup_method")
        self.add_line("self.backup_queue = backup_queue")
        self.add_line("self.backup_url = backup_url")
        self.add_line("self.backup_period = backup_period")
        self.add_line("self.backup_delay = backup_delay")
        self.add_line('self.connections = kwargs.get("connections", [])')
        self.add_line('self.msgRCount = self.limit_number(kwargs.get("msgRCount", 0))')
        self.add_line('self.msgSCount = self.limit_number(kwargs.get("msgSCount", 0))')

        for init_float_param in agent.init_floats.values():
            name = init_float_param.name
            value = init_float_param.value
            self.add_line(
                f'self.{name} = self.limit_number(kwargs.get("{name}", {value}))'
            )

        for dist_normal_float_param in agent.dist_normal_floats.values():
            name = dist_normal_float_param.name
            mean = f"self.limit_number({dist_normal_float_param.mean})"
            std_dev = f"self.limit_number({dist_normal_float_param.std_dev})"
            self.add_line(
                f'self.{name} = self.limit_number(kwargs.get("{name}", numpy.random.normal({mean}, {std_dev})))'
            )

        for dist_exp_float_param in agent.dist_exp_floats.values():
            name = dist_exp_float_param.name
            lambda_ = f"self.limit_number({dist_exp_float_param.lambda_})"
            self.add_line(
                f'self.{name} = self.limit_number(kwargs.get("{name}", numpy.random.exponential(self.limit_number(1 / {lambda_}))))'
            )

        for dist_uniform_float_param in agent.dist_unifrom_floats.values():
            name = dist_uniform_float_param.name
            a = f"self.limit_number({dist_uniform_float_param.a})"
            b = f"self.limit_number({dist_uniform_float_param.b})"
            self.add_line(
                f'self.{name} = self.limit_number(kwargs.get("{name}", random.uniform({a}, {b})))'
            )

        for enum_param in agent.enums.values():
            value_list: List[str] = []
            percentage_list: List[str] = []
            for enum_value in enum_param.enum_values:
                value_list.append(f'"{enum_value.value}"')
                percentage_list.append(enum_value.percentage)
            values = f'[{", ".join(value_list)}]'
            percentages = f'[{", ".join(percentage_list)}]'
            self.add_line(
                f'self.{enum_param.name} = kwargs.get("{enum_param.name}", random.choices({values}, {percentages})[0])'
            )

        for connection_list_param in agent.connection_lists.values():
            self.add_line(
                f'self.{connection_list_param.name} = kwargs.get("{connection_list_param.name}", [])'
            )

        for message_list_param in agent.message_lists.values():
            self.add_line(
                f'self.{message_list_param.name} = kwargs.get("{message_list_param.name}", [])'
            )

        for float_list_param in agent.float_lists.values():
            self.add_line(
                f'self.{float_list_param.name} = kwargs.get("{float_list_param.name}", [])'
            )

        self.add_line(
            "if self.logger: self.logger.debug(f'[{self.jid}] Class dict after initialization: {self.__dict__}')"
        )

        self.indent_left()
        self.add_newline()

        self.add_line("@property")
        self.add_line("def connCount(self):")
        self.indent_right()
        self.add_line("return self.limit_number(len(self.connections))")

        self.indent_left()

    def add_float_utils(self) -> None:
        self.add_line("def limit_number(self, value):")
        self.indent_right()
        self.add_line("return float(max(-2147483648, min(value, 2147483647)))")
        self.indent_left()

    def add_message_utils(self) -> None:
        self.add_line("def get_json_from_spade_message(self, msg):")
        self.indent_right()
        self.add_line("return orjson.loads(msg.body)")
        self.indent_left()
        self.add_newline()

        self.add_line("def get_spade_message(self, receiver_jid, body):")
        self.indent_right()
        self.add_line("msg = spade.message.Message(to=receiver_jid)")
        self.add_line('body["sender"] = str(self.jid)')
        self.add_line('msg.metadata["type"] = body["type"]')
        self.add_line('msg.metadata["performative"] = body["performative"]')
        self.add_line('msg.body = str(orjson.dumps(body), encoding="utf-8")')
        self.add_line("return msg")

        self.indent_left()

    def add_agent_setup(self, agent: Agent) -> None:
        self.add_line("def setup(self):")
        self.indent_right()

        self.add_line("if self.backup_method is not None:")
        self.indent_right()
        self.add_no_match_template("BackupBehaviour")
        self.add_line(
            "self.add_behaviour(self.BackupBehaviour(start_at=datetime.datetime.now() + datetime.timedelta(seconds=self.backup_delay), period=self.backup_period), BackupBehaviour_template)"
        )
        self.indent_left()

        for setup_behaviour in agent.setup_behaviours.values():
            self.add_no_match_template(f"{setup_behaviour.name}")
            self.add_line(
                f"self.add_behaviour(self.{setup_behaviour.name}(), {setup_behaviour.name}_template)"
            )

        for one_time_behaviour in agent.one_time_behaviours.values():
            self.add_no_match_template(f"{one_time_behaviour.name}")
            self.add_line(
                f"self.add_behaviour(self.{one_time_behaviour.name}(start_at=datetime.datetime.now() + datetime.timedelta(seconds={one_time_behaviour.delay})), {one_time_behaviour.name}_template)"
            )

        for cyclic_behaviour in agent.cyclic_behaviours.values():
            self.add_no_match_template(f"{cyclic_behaviour.name}")
            self.add_line(
                f"self.add_behaviour(self.{cyclic_behaviour.name}(period={cyclic_behaviour.period}), {cyclic_behaviour.name}_template)"
            )

        for message_received_behaviour in agent.message_received_behaviours.values():
            self.add_line(
                f"{message_received_behaviour.name}_template = spade.template.Template()"
            )
            self.add_line(
                f'{message_received_behaviour.name}_template.set_metadata("type", "{message_received_behaviour.received_message.type}")'
            )
            self.add_line(
                f'{message_received_behaviour.name}_template.set_metadata("performative", "{message_received_behaviour.received_message.performative}")'
            )
            self.add_line(
                f"self.add_behaviour(self.{message_received_behaviour.name}(), {message_received_behaviour.name}_template)"
            )

        self.add_line(
            "if self.logger: self.logger.debug(f'[{self.jid}] Class dict after setup: {self.__dict__}')"
        )

        self.indent_left()

    def add_no_match_template(self, behaviour_name: str) -> None:
        self.add_line(f"{behaviour_name}_template = spade.template.Template()")
        self.add_line(
            f'{behaviour_name}_template.set_metadata("reserved", "no_message_match")'
        )

    def add_backup_behaviour(self, agent: Agent) -> None:
        self.add_line("class BackupBehaviour(spade.behaviour.PeriodicBehaviour):")
        self.indent_right()
        self.add_line("def __init__(self, start_at, period):")
        self.indent_right()
        self.add_line("super().__init__(start_at=start_at, period=period)")
        self.add_line("self.http_client = httpx.AsyncClient(timeout=period)")
        self.indent_left()
        self.add_newline()

        self.add_line("async def run(self):")
        self.indent_right()
        self.add_line("data = {")
        self.indent_right()
        self.add_line(
            '"__timestamp__": int(datetime.datetime.timestamp(datetime.datetime.utcnow())),'
        )
        self.add_line('"jid": str(self.agent.jid),')
        self.add_line(f'"type": "{agent.name}",')

        self.add_line('"floats": {')
        self.indent_right()
        self.add_line('"msgRCount": self.agent.msgRCount,')
        self.add_line('"msgSCount": self.agent.msgSCount,')
        self.add_line('"connCount": self.agent.connCount,')
        for float_param_name in agent.float_param_names:
            self.add_line(f'"{float_param_name}": self.agent.{float_param_name},')
        self.indent_left()
        self.add_line("},")

        self.add_line('"enums": {')
        self.indent_right()
        for enum_param_name in agent.enums:
            self.add_line(f'"{enum_param_name}": self.agent.{enum_param_name},')
        self.indent_left()
        self.add_line("},")

        self.add_line('"connections": {')
        self.indent_right()
        self.add_line('"connections": self.agent.connections,')
        for connection_list_param_name in agent.connection_lists:
            self.add_line(
                f'"{connection_list_param_name}": self.agent.{connection_list_param_name},'
            )
        self.indent_left()
        self.add_line("},")

        self.add_line('"messages": {')
        self.indent_right()
        for message_list_param_name in agent.message_lists:
            self.add_line(
                f'"{message_list_param_name}": self.agent.{message_list_param_name},'
            )
        self.indent_left()
        self.add_line("},")

        self.add_line('"float_lists": {')
        self.indent_right()
        for float_list_param_name in agent.float_lists:
            self.add_line(
                f'"{float_list_param_name}": self.agent.{float_list_param_name},'
            )
        self.indent_left()
        self.add_line("},")
        self.indent_left()
        self.add_line("}")

        self.add_line("if self.agent.backup_method == 'http':")
        self.indent_right()
        self.add_line(
            "if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Sending backup data with http: {data}')"
        )
        self.add_line("try:")
        self.indent_right()
        self.add_line(
            'await self.http_client.post(self.agent.backup_url, headers={"Content-Type": "application/json"}, data=orjson.dumps(data))'
        )
        self.indent_left()
        self.add_line("except Exception as e:")
        self.indent_right()
        self.add_line(
            "if self.agent.logger: self.agent.logger.error(f'[{self.agent.jid}] Backup error type: {e.__class__}, additional info: {e}')"
        )
        self.indent_left()
        self.indent_left()
        self.add_line("elif self.agent.backup_method == 'queue':")
        self.indent_right()
        self.add_line(
            "if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Sending backup data with queue: {data}')"
        )
        self.add_line("try:")
        self.indent_right()
        self.add_line("await self.agent.backup_queue.coro_put(data)")
        self.indent_left()
        self.add_line("except Exception as e:")
        self.indent_right()
        self.add_line(
            "if self.agent.logger: self.agent.logger.error(f'[{self.agent.jid}] Backup error type: {e.__class__}, additional info: {e}')"
        )
        self.indent_left()
        self.indent_left()
        self.add_line("else:")
        self.indent_right()
        self.add_line(
            "if self.agent.logger: self.agent.logger.warning(f'[{self.agent.jid}] Unknown backup method: {self.agent.backup_method}')"
        )
        self.indent_left()
        self.indent_left()
        self.indent_left()

    def add_agent_behaviour(self, behaviour: Behaviour, behaviour_type: str) -> None:
        self.add_line(f"class {behaviour.name}({behaviour_type}):")
        self.indent_right()

        for action in behaviour.actions.values():
            self.add_action(behaviour, action)

        self.add_line("async def run(self):")
        self.indent_right()
        if isinstance(behaviour, MessageReceivedBehaviour):
            self.add_rcv_message()
        elif not behaviour.actions.values():
            self.add_line("...")
            self.indent_left()
            self.indent_left()
            return

        if isinstance(behaviour, MessageReceivedBehaviour):
            self.indent_right()

        for action in behaviour.actions.values():
            self.add_action_call(behaviour, action)

        if isinstance(behaviour, MessageReceivedBehaviour):
            self.indent_left()

        self.indent_left()
        self.indent_left()

    def add_rcv_message(self) -> None:
        self.add_line("rcv = await self.receive(timeout=100000)")
        self.add_line("if rcv:")
        self.indent_right()
        self.add_line("rcv = self.agent.get_json_from_spade_message(rcv)")
        self.add_line(
            "self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)"
        )
        self.add_line(
            "if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')"
        )
        self.indent_left()

    def add_action_call(self, behaviour: Behaviour, action: Action) -> None:
        action_call = ""
        if isinstance(action, SendMessageAction):
            action_call += "await "
        action_call += f"self.{action.name}"
        if isinstance(behaviour, MessageReceivedBehaviour):
            action_call += "(rcv)"
        else:
            action_call += "()"
        self.add_line(action_call)

    def add_action(self, behaviour: Behaviour, action: Action) -> None:
        self.add_action_def(behaviour, action)
        self.indent_right()

        self.add_line(
            f"if self.agent.logger: self.agent.logger.debug(f'[{{self.agent.jid}}] Run action {action.name}')"
        )

        if isinstance(action, SendMessageAction):
            self.add_send_message(action.send_message)

        self.add_block(action.main_block)
        self.indent_left()

        self.add_newline()

    def add_action_def(self, behaviour: Behaviour, action: Action) -> None:
        action_def = ""
        if isinstance(action, SendMessageAction):
            action_def += "async "
        action_def += f"def {action.name}"
        if isinstance(behaviour, MessageReceivedBehaviour):
            action_def += "(self, rcv):"
        else:
            action_def += "(self):"
        self.add_line(action_def)

    def add_send_message(self, message: IntermediateMessage) -> None:
        send_msg = f'send = {{ "type": "{message.type}", "performative": "{message.performative}", '
        for float_param_name in message.float_params:
            send_msg += f'"{float_param_name}": 0.0, '
        for connection_param_name in message.connection_params:
            send_msg += f'"{connection_param_name}": "", '
        send_msg += "}"
        self.add_line(send_msg)

    def parse_arg(self, arg: Argument) -> str:
        match arg.type_in_op:
            case AgentParam():
                return f"self.agent.{arg.expr}"

            case EnumValue():
                return f'"{arg.expr}"'

            case ReceivedMessageParam():
                prop = arg.expr.split(".")[1]
                return f'rcv["{prop}"]'

            case SendMessageParam():
                prop = arg.expr.split(".")[1]
                return f'send["{prop}"]'

            case _:
                return arg.expr

    def add_block(self, block: Block) -> None:
        if not block.statements:
            self.add_line("...")
            return

        for statement in block.statements:
            match statement:
                case Block():
                    self.indent_right()
                    self.add_block(statement)
                    self.indent_left()

                case FloatDeclaration():
                    self.add_line("")
                    self.add_line("# float declaration")
                    value = (
                        f"self.agent.limit_number({self.parse_arg(statement.value)})"
                    )
                    self.add_line(f"{statement.name} = {value}")

                case ConnectionDeclaration():
                    self.add_line("")
                    self.add_line("# connection declaration")
                    value = f"{self.parse_arg(statement.value)}"
                    self.add_line(f"{statement.name} = {value}")

                case Subset():
                    self.add_line("")
                    self.add_line("# subset")
                    dst_list = self.parse_arg(statement.dst_list)
                    src_list = self.parse_arg(statement.src_list)
                    src_list_len = f"int(self.agent.limit_number(len({src_list})))"
                    num = f"int(self.agent.limit_number(round(self.agent.limit_number({self.parse_arg(statement.num)}))))"
                    self.add_line(f"if {num} <= 0:")
                    self.indent_right()
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.debug(f'[{{self.agent.jid}}] Non-positive subset size (rounded): \u007b{num}\u007d')"
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(
                        f"{dst_list} = [copy.deepcopy(elem) for elem in random.sample({src_list}, min({num}, {src_list_len}))]"
                    )

                case Clear():
                    self.add_line("")
                    self.add_line("# clear")
                    list_ = self.parse_arg(statement.list_)
                    self.add_line(f"{list_}.clear()")

                case Send() if isinstance(statement.receivers.type_in_op, Connection):
                    self.add_line("")
                    self.add_line("# send")
                    receiver = self.parse_arg(statement.receivers)
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.debug(f'[{{self.agent.jid}}] Send message {{send}} to \u007b{receiver}\u007d')"
                    )
                    self.add_line(
                        f"await self.send(self.agent.get_spade_message({receiver}, send))"
                    )
                    self.add_line(
                        "self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)"
                    )

                case Send() if isinstance(
                    statement.receivers.type_in_op, ConnectionList
                ):
                    self.add_line("")
                    self.add_line("# send")
                    receivers = self.parse_arg(statement.receivers)
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.debug(f'[{{self.agent.jid}}] Send message {{send}} to \u007b{receivers}\u007d')"
                    )
                    self.add_line(f"for receiver in {receivers}:")
                    self.indent_right()
                    self.add_line(
                        "await self.send(self.agent.get_spade_message(receiver, send))"
                    )
                    self.add_line(
                        "self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)"
                    )
                    self.indent_left()

                case Set() if isinstance(statement.value.type_in_op, MessageList):
                    self.add_line("")
                    self.add_line("# set")
                    msg = self.parse_arg(statement.dst)
                    msg_list = self.parse_arg(statement.value)
                    self.add_line(
                        f'if not any([msg["type"] == {msg}["type"] and msg["performative"] == {msg}["performative"] for msg in {msg_list}]):'
                    )
                    self.indent_right()
                    self.add_line(
                        f'if self.agent.logger: self.agent.logger.warning(f\'[{{self.agent.jid}}] No messages with type/performative \u007b{msg}["type"]\u007d/\u007b{msg}["performative"]\u007d found in list {msg_list}\')'
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(
                        f'{msg} = copy.deepcopy(random.choice(list(filter(lambda msg: msg["type"] == {msg}["type"] and msg["performative"] == {msg}["performative"], {msg_list}))))'
                    )

                case Set() if isinstance(statement.value.type_in_op, Float):
                    self.add_line("")
                    self.add_line("# set")
                    dst = self.parse_arg(statement.dst)
                    num = f"self.agent.limit_number({self.parse_arg(statement.value)})"
                    self.add_line(f"{dst} = {num}")

                case Set():
                    self.add_line("")
                    self.add_line("# set")
                    dst = self.parse_arg(statement.dst)
                    value = self.parse_arg(statement.value)
                    self.add_line(f"{dst} = {value}")

                case Round():
                    self.add_line("")
                    self.add_line("# round")
                    dst = self.parse_arg(statement.dst)
                    self.add_line(
                        f"{dst} = self.agent.limit_number(round(self.agent.limit_number({dst})))"
                    )

                case UniformDist():
                    self.add_line("")
                    self.add_line("# uniform distribution")
                    dst = self.parse_arg(statement.dst)
                    a = f"self.agent.limit_number({self.parse_arg(statement.a)})"
                    b = f"self.agent.limit_number({self.parse_arg(statement.b)})"
                    self.add_line(
                        f"{dst} = self.agent.limit_number(random.uniform({a}, {b}))"
                    )

                case NormalDist():
                    self.add_line("")
                    self.add_line("# normal distribution")
                    dst = self.parse_arg(statement.dst)
                    mean = f"self.agent.limit_number({self.parse_arg(statement.mean)})"
                    std_dev = (
                        f"self.agent.limit_number({self.parse_arg(statement.std_dev)})"
                    )
                    self.add_line(f"if {std_dev} < 0:")
                    self.indent_right()
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.warning(f'[{{self.agent.jid}}] Negative standard deviation: \u007b{std_dev}\u007d')"
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(
                        f"{dst} = self.agent.limit_number(numpy.random.normal({mean}, {std_dev}))"
                    )

                case ExpDist():
                    self.add_line("")
                    self.add_line("# exponential distribution")
                    dst = self.parse_arg(statement.dst)
                    lambda_ = (
                        f"self.agent.limit_number({self.parse_arg(statement.lambda_)})"
                    )
                    self.add_line(f"if {lambda_} <= 0:")
                    self.indent_right()
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.warning(f'[{{self.agent.jid}}] Non-positive lambda: \u007b{lambda_}\u007d')"
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(
                        f"{dst} = self.agent.limit_number(numpy.random.exponential(self.agent.limit_number(1 / {lambda_})))"
                    )

                case Comparaison():
                    if isinstance(statement.left.type_in_op, Float):
                        left = (
                            f"self.agent.limit_number({self.parse_arg(statement.left)})"
                        )
                    else:
                        left = self.parse_arg(statement.left)

                    if isinstance(statement.right.type_in_op, Float):
                        right = f"self.agent.limit_number({self.parse_arg(statement.right)})"
                    else:
                        right = self.parse_arg(statement.right)

                    match statement:
                        case IfGreaterThan():
                            self.add_line("")
                            self.add_line("# if greater than")
                            self.add_line(f"if {left} > {right}:")

                        case IfGreaterThanOrEqual():
                            self.add_line("")
                            self.add_line("# if greater than or equal")
                            self.add_line(f"if {left} >= {right}:")

                        case IfLessThan():
                            self.add_line("")
                            self.add_line("# if less than")
                            self.add_line(f"if {left} < {right}:")

                        case IfLessThanOrEqual():
                            self.add_line("")
                            self.add_line("# if greater than or equal")
                            self.add_line(f"if {left} <= {right}:")

                        case IfEqual():
                            self.add_line("")
                            self.add_line("# if equal")
                            self.add_line(f"if {left} == {right}:")

                        case IfNotEqual():
                            self.add_line("")
                            self.add_line("# if not equal")
                            self.add_line(f"if {left} != {right}:")

                        case WhileGreaterThan():
                            self.add_line("")
                            self.add_line("# while greater than")
                            self.add_line(f"while {left} > {right}:")

                        case WhileGreaterThanOrEqual():
                            self.add_line("")
                            self.add_line("# while greater than or equal")
                            self.add_line(f"while {left} >= {right}:")

                        case WhileLessThan():
                            self.add_line("")
                            self.add_line("# while less than")
                            self.add_line(f"while {left} < {right}:")

                        case WhileLessThanOrEqual():
                            self.add_line("")
                            self.add_line("# while less than or equal")
                            self.add_line(f"while {left} <= {right}:")

                        case WhileEqual():
                            self.add_line("")
                            self.add_line("# while equal")
                            self.add_line(f"while {left} == {right}:")

                        case WhileNotEqual():
                            self.add_line("")
                            self.add_line("# while not equal")
                            self.add_line(f"while {left} != {right}:")

                        case _:
                            raise Exception(
                                f"Unknown comparaison statement: {statement.print()}"
                            )

                case MathOperation():
                    dst = self.parse_arg(statement.dst)
                    num = f"self.agent.limit_number({self.parse_arg(statement.num)})"
                    match statement:
                        case Add():
                            self.add_line("")
                            self.add_line("# add")
                            self.add_line(
                                f"{dst} = self.agent.limit_number({dst} + {num})"
                            )

                        case Subtract():
                            self.add_line("")
                            self.add_line("# subtract")
                            self.add_line(
                                f"{dst} = self.agent.limit_number({dst} - {num})"
                            )

                        case Multiply():
                            self.add_line("")
                            self.add_line("# multiply")
                            self.add_line(
                                f"{dst} = self.agent.limit_number({dst} * {num})"
                            )

                        case Divide():
                            self.add_line("")
                            self.add_line("# divide")
                            self.add_line(f"if {num} == 0:")
                            self.indent_right()
                            self.add_line(
                                f"if self.agent.logger: self.agent.logger.warning(f'[{{self.agent.jid}}] Division by zero: \u007b{num}\u007d')"
                            )
                            self.add_line("return")
                            self.indent_left()
                            self.add_line(
                                f"{dst} = self.agent.limit_number({dst} / {num})"
                            )

                        case _:
                            raise Exception(
                                f"Unknown math operation statement: {statement.print()}"
                            )

                case Modulo():
                    dst = self.parse_arg(statement.dst)
                    dividend = (
                        f"self.agent.limit_number({self.parse_arg(statement.dividend)})"
                    )
                    divisor = (
                        f"self.agent.limit_number({self.parse_arg(statement.divisor)})"
                    )
                    self.add_line("")
                    self.add_line("# modulo")
                    self.add_line(f"if {divisor} == 0:")
                    self.indent_right()
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.warning(f'[{{self.agent.jid}}] Modulo division by zero: \u007b{divisor}\u007d')"
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(
                        f"{dst} = self.agent.limit_number({dividend} % {divisor})"
                    )

                case ListElementAccess():
                    list_ = self.parse_arg(statement.list_)
                    element = self.parse_arg(statement.element)
                    match statement:
                        case AddElement() if isinstance(
                            statement.element.type_in_op, MessageList
                        ):
                            self.add_line("")
                            self.add_line("# add element")
                            self.add_line(
                                f"if {element} not in {list_}: {list_}.append({element})"
                            )

                        case AddElement():
                            self.add_line("")
                            self.add_line("# add element")
                            self.add_line(f"{list_}.append({element})")

                        case RemoveElement():
                            self.add_line("")
                            self.add_line("# remove element")
                            self.add_line(
                                f"if {element} in {list_}: {list_}.remove({element})"
                            )

                        case IfInList():
                            self.add_line("")
                            self.add_line("# if in list")
                            self.add_line(f"if {element} in {list_}:")

                        case IfNotInList():
                            self.add_line("")
                            self.add_line("# if not in list")
                            self.add_line(f"if {element} not in {list_}:")

                        case _:
                            raise Exception(
                                f"Unknown list element access statement: {statement.print()}"
                            )

                case RemoveNElements():
                    self.add_line("")
                    self.add_line("# remove n elements")
                    list_ = self.parse_arg(statement.list_)
                    list_len = f"int(self.agent.limit_number(len({list_})))"
                    num = f"int(self.agent.limit_number(round(self.agent.limit_number({self.parse_arg(statement.num)}))))"
                    self.add_line(f"if {num} < 0 or {num} > {list_len}:")
                    self.indent_right()
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.debug(f'[{{self.agent.jid}}] Incorrect number of elements to remove (rounded, either negative or bigger than the list size): \u007b{num}\u007d')"
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(f"random.shuffle({list_})")
                    self.add_line(
                        f"{list_} = {list_}[:int(self.agent.limit_number({list_len} - {num}))]"
                    )

                case Length():
                    self.add_line("")
                    self.add_line("# length")
                    dst = self.parse_arg(statement.dst)
                    list_ = self.parse_arg(statement.list_)
                    self.add_line(f"{dst} = self.agent.limit_number(len({list_}))")

                case ListRead():
                    self.add_line("")
                    self.add_line("# list read")
                    dst = self.parse_arg(statement.dst)
                    list_ = self.parse_arg(statement.list_)
                    idx = f"int(self.agent.limit_number(round(self.agent.limit_number({self.parse_arg(statement.idx)}))))"
                    match statement.list_.type_in_op:
                        case FloatList():
                            value = f"self.agent.limit_number({list_}[{idx}])"
                        case _:
                            value = f"{list_}[{idx}]"
                    list_len = f"int(self.agent.limit_number(len({list_})))"
                    self.add_line(f"if {idx} < 0 or {idx} >= {list_len}:")
                    self.indent_right()
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.warning(f'[{{self.agent.jid}}] Incorrect index (rounded, either negative or bigger than the list size): \u007b{idx}\u007d')"
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(f"{dst} = {value}")

                case ListWrite():
                    self.add_line("")
                    self.add_line("# list write")
                    list_ = self.parse_arg(statement.list_)
                    idx = f"int(self.agent.limit_number(round(self.agent.limit_number({self.parse_arg(statement.idx)}))))"
                    match statement.value.type_in_op:
                        case Float():
                            value = f"self.agent.limit_number({self.parse_arg(statement.value)})"
                        case _:
                            value = self.parse_arg(statement.value)
                    dst = f"{list_}[{idx}]"
                    list_len = f"int(self.agent.limit_number(len({list_})))"
                    self.add_line(f"if {idx} < 0 or {idx} >= {list_len}:")
                    self.indent_right()
                    self.add_line(
                        f"if self.agent.logger: self.agent.logger.warning(f'[{{self.agent.jid}}] Incorrect index (rounded, either negative or bigger than the list size): \u007b{idx}\u007d')"
                    )
                    self.add_line("return")
                    self.indent_left()
                    self.add_line(f"{dst} = {value}")

                case TrigonometryOperation():
                    dst = self.parse_arg(statement.dst)
                    degree = (
                        f"self.agent.limit_number({self.parse_arg(statement.degree)})"
                    )

                    match statement:
                        case Sin():
                            self.add_line("")
                            self.add_line("# sin")
                            self.add_line(
                                f"{dst} = self.agent.limit_number(numpy.sin(numpy.deg2rad({degree})))"
                            )

                        case Cos():
                            self.add_line("")
                            self.add_line("# cos")
                            self.add_line(
                                f"{dst} = self.agent.limit_number(numpy.cos(numpy.deg2rad({degree})))"
                            )

                        case _:
                            raise Exception(
                                f"Unknown trigonometry operation statement: {statement.print()}"
                            )

                case ExponentiationOperation():
                    dst = self.parse_arg(statement.dst)
                    base = f"self.agent.limit_number({self.parse_arg(statement.base)})"
                    num = f"self.agent.limit_number({self.parse_arg(statement.num)})"

                    match statement:
                        case Power():
                            self.add_line("")
                            self.add_line("# power")
                            self.add_line(
                                f"{dst} = self.agent.limit_number(numpy.power({base}, {num}))"
                            )

                        case Logarithm():
                            self.add_line("")
                            self.add_line("# logarithm")
                            self.add_line(
                                f"if {base} <= 0 or {num} <= 0 or {base} == 1:"
                            )
                            self.indent_right()
                            self.add_line(
                                f"if self.agent.logger: self.agent.logger.warning(f'[{{self.agent.jid}}] Incorrect logarithm arguments: log(\u007b{base}\u007d, \u007b{num}\u007d')"
                            )
                            self.add_line("return")
                            self.indent_left()
                            numerator = f"self.agent.limit_number(numpy.log({num}))"
                            denominator = f"self.agent.limit_number(numpy.log({base}))"
                            self.add_line(
                                f"{dst} = self.agent.limit_number({numerator} / {denominator})"
                            )

                        case _:
                            raise Exception(
                                f"Unknown exponentiation operation statement: {statement.print()}"
                            )

                case _:
                    raise Exception(f"Unknown statement: {statement.print()}")



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.utils.validation import is_int

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_SIZE(state: State, size: str) -> None:
    state.require(
        state.in_graph, "Not inside any graph.", "SIZE can be used inside graphs."
    )
    state.require(not state.last_graph.is_size_defined(), "Size is already defined")
    state.require(is_int(size), "Graph size must be an integer value.")
    state.require(int(size) >= 0, "Graph size cannot be a negative number.")

    state.last_graph.set_size(int(size))



from __future__ import annotations

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from aasm.intermediate.declaration import Declaration
    from aasm.intermediate.instruction import Instruction


class Declarations:
    def __init__(
        self,
        float_names: List[str] | None = None,
        connection_names: List[str] | None = None,
    ):
        if float_names:
            self.float_names: List[str] = float_names
        else:
            self.float_names: List[str] = []

        if connection_names:
            self.connection_names: List[str] = connection_names
        else:
            self.connection_names: List[str] = []

    def add_float_name(self, name: str) -> None:
        self.float_names.append(name)

    def is_float_name(self, name: str) -> bool:
        return name in self.float_names

    def add_connection_name(self, name: str) -> None:
        self.connection_names.append(name)

    def is_connection_name(self, name: str) -> bool:
        return name in self.connection_names

    def get_declared_names(self) -> List[str]:
        return [*self.float_names, *self.connection_names]

    def get_copy(self) -> Declarations:
        return Declarations(list(self.float_names), list(self.connection_names))

    def print(self) -> None:
        print("Declarations")
        print(f"float_names = {self.float_names}")
        print(f"connection_names = {self.connection_names}")


class Block:
    def __init__(self, parent_declarations: Declarations):
        self.statements: List[Declaration | Instruction | Block] = []
        self._declarations: Declarations = parent_declarations.get_copy()

    @property
    def declared_names_in_scope(self) -> List[str]:
        return self._declarations.get_declared_names()

    @property
    def get_declarations(self) -> Declarations:
        return self._declarations

    def is_declared_float(self, name: str) -> bool:
        return self._declarations.is_float_name(name)

    def is_declared_connection(self, name: str) -> bool:
        return self._declarations.is_connection_name(name)

    def add_float_declaration(self, declaration: Declaration) -> None:
        self._declarations.add_float_name(declaration.name)
        self.statements.append(declaration)

    def add_connection_declaration(self, declaration: Declaration) -> None:
        self._declarations.add_connection_name(declaration.name)
        self.statements.append(declaration)

    def add_statement(self, statement: Instruction | Block) -> None:
        self.statements.append(statement)

    def print(self) -> None:
        print(f"Block")
        print(f"Names in scope")
        self._declarations.print()
        for instruction in self.statements:
            instruction.print()
        print("(EndBlock)")



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.behaviour import (
    CyclicBehaviour,
    MessageReceivedBehaviour,
    OneTimeBehaviour,
    SetupBehaviour,
)
from aasm.utils.validation import is_float, is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_BEHAV(state: State, name: str, category: str, args: List[str]):
    state.require(
        state.in_agent,
        "Cannot define behaviours outside agents.",
        "Try defining new agents using AGENT.",
    )
    state.require(
        not state.in_behaviour,
        "Cannot define behaviours inside other behaviours.",
        "First end current behaviour using EBEHAV.",
    )
    state.require(
        not state.last_agent.name_exists(name),
        f"{name} already exists in current agent.",
    )
    state.require(
        is_valid_name(name),
        f"{name} is not a correct name.",
        f"Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.",
    )

    match category, args:
        case "setup", []:
            state.last_agent.add_setup_behaviour(SetupBehaviour(name))

        case "one_time", [delay]:
            state.require(
                is_float(delay) and float(delay) >= 0,
                f"{delay} is not a valid delay value.",
            )

            state.last_agent.add_one_time_behaviour(OneTimeBehaviour(name, delay))

        case "cyclic", [period]:
            state.require(
                is_float(period) and float(period) > 0,
                f"{period} is not a valid period value.",
            )

            state.last_agent.add_cyclic_behaviour(CyclicBehaviour(name, period))

        case "msg_rcv", [msg_type, msg_performative]:
            state.require(
                state.message_exists(msg_type, msg_performative),
                f"Message {msg_type}/{msg_performative} does not exist.",
                "Try defining new messages using MESSAGE",
            )
            state.require(
                not state.last_agent.behaviour_for_template_exists(
                    msg_type, msg_performative
                ),
                f"Behaviour for template {msg_type}/{msg_performative} already exists in current agent.",
            )

            state.last_agent.add_message_received_behaviour(
                MessageReceivedBehaviour(
                    name, state.get_message_instance(msg_type, msg_performative)
                )
            )

        case _:
            state.panic(f"Incorrect operation: BEHAV {name} {category} {args}")

    state.in_behaviour = True


def op_EBEHAV(state: State) -> None:
    state.require(
        state.in_behaviour,
        "Not inside any behaviour.",
        "Try defining new behaviours using BEHAV.",
    )
    state.require(
        not state.in_action,
        "Cannot end behaviours inside actions.",
        "Try ending current action using EACTION.",
    )

    state.in_behaviour = False



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.utils.validation import is_int

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_SCALE(state: State, scale: str) -> None:
    state.require(
        state.in_graph,
        "Not inside any graph",
        "Try defining new graph using GRAPH.",
    )
    state.require(not state.last_graph.is_scale_defined(), "Scale is already defined")
    state.require(is_int(scale), "Scale must be an integer")
    state.require(int(scale) > 0, "Scale must be greater than 0")
    state.last_graph.set_scale(int(scale))



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aasm.intermediate.argument import Argument


class Instruction:
    def print(self) -> None:
        raise NotImplementedError()


class Comparaison(Instruction):
    def __init__(self, left: Argument, right: Argument):
        self.left = left
        self.right = right

    def print(self) -> None:
        print("Comparaison")
        self.left.print()
        self.right.print()


class IfGreaterThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("IfGreaterThan")
        super().print()


class IfGreaterThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("IfGreaterThanOrEqual")
        super().print()


class IfLessThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("IfLessThan")
        super().print()


class IfLessThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("IfLessThanOrEqual")
        super().print()


class IfEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("IfEqual")
        super().print()


class IfNotEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("IfNotEqual")
        super().print()


class WhileGreaterThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("WhileGreaterThan")
        super().print()


class WhileGreaterThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("WhileGreaterThanOrEqual")
        super().print()


class WhileLessThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("WhileLessThan")
        super().print()


class WhileLessThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("WhileLessThanOrEqual")
        super().print()


class WhileEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("WhileEqual")
        super().print()


class WhileNotEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print("WhileNotEqual")
        super().print()


class MathOperation(Instruction):
    def __init__(self, dst: Argument, num: Argument):
        self.dst = dst
        self.num = num

    def print(self) -> None:
        print("MathOperation")
        self.dst.print()
        self.num.print()


class Multiply(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print("Multiply")
        super().print()


class Divide(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print("Divide")
        super().print()


class Add(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print("Add")
        super().print()


class Subtract(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print("Subtract")
        super().print()


class ListElementAccess(Instruction):
    def __init__(self, list_: Argument, element: Argument):
        self.list_ = list_
        self.element = element

    def print(self) -> None:
        print("ListElementAccess")
        self.list_.print()
        self.element.print()


class AddElement(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print("AddElement")
        super().print()


class RemoveElement(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print("RemoveElement")
        super().print()


class IfInList(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print("IfInList")
        super().print()


class IfNotInList(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print("IfNotInList")
        super().print()


class RemoveNElements(Instruction):
    def __init__(self, list_: Argument, num: Argument):
        self.list_ = list_
        self.num = num

    def print(self) -> None:
        print("RemoveNElements")
        self.list_.print()
        self.num.print()


class Subset(Instruction):
    def __init__(self, dst_list: Argument, src_list: Argument, num: Argument):
        self.dst_list = dst_list
        self.src_list = src_list
        self.num = num

    def print(self) -> None:
        print("Subset")
        self.dst_list.print()
        self.src_list.print()
        self.num.print()


class Set(Instruction):
    def __init__(self, dst: Argument, value: Argument):
        self.dst = dst
        self.value = value

    def print(self) -> None:
        print("Set")
        self.dst.print()
        self.value.print()


class Clear(Instruction):
    def __init__(self, list_: Argument):
        self.list_ = list_

    def print(self) -> None:
        print("Clear")
        self.list_.print()


class Length(Instruction):
    def __init__(self, dst: Argument, list_: Argument):
        self.dst = dst
        self.list_ = list_

    def print(self) -> None:
        print("Length")
        self.dst.print()
        self.list_.print()


class Send(Instruction):
    def __init__(self, receivers: Argument):
        self.receivers = receivers

    def print(self) -> None:
        print("Send")
        self.receivers.print()


class Round(Instruction):
    def __init__(self, dst: Argument):
        self.dst = dst

    def print(self) -> None:
        print("Round")
        self.dst.print()


class UniformDist(Instruction):
    def __init__(self, dst: Argument, a: Argument, b: Argument):
        self.dst = dst
        self.a = a
        self.b = b

    def print(self) -> None:
        print("UniformDist")
        self.dst.print()
        self.a.print()
        self.b.print()


class NormalDist(Instruction):
    def __init__(self, dst: Argument, mean: Argument, std_dev: Argument):
        self.dst = dst
        self.mean = mean
        self.std_dev = std_dev

    def print(self) -> None:
        print("NormalDist")
        self.dst.print()
        self.mean.print()
        self.std_dev.print()


class ExpDist(Instruction):
    def __init__(self, dst: Argument, lambda_: Argument):
        self.dst = dst
        self.lambda_ = lambda_

    def print(self) -> None:
        print("ExpDist")
        self.dst.print()
        self.lambda_.print()


class ListRead(Instruction):
    def __init__(self, dst: Argument, list_: Argument, idx: Argument):
        self.dst = dst
        self.list_ = list_
        self.idx = idx

    def print(self) -> None:
        print("ListRead")
        self.dst.print()
        self.list_.print()
        self.idx.print()


class ListWrite(Instruction):
    def __init__(self, list_: Argument, idx: Argument, value: Argument):
        self.list_ = list_
        self.idx = idx
        self.value = value

    def print(self) -> None:
        print("ListWrite")
        self.list_.print()
        self.idx.print()
        self.value.print()


class TrigonometryOperation(Instruction):
    def __init__(self, dst: Argument, degree: Argument):
        self.dst = dst
        self.degree = degree

    def print(self) -> None:
        print("TrigonometryOperation")
        self.dst.print()
        self.degree.print()


class Sin(TrigonometryOperation):
    def __init__(self, dst: Argument, degree: Argument):
        super().__init__(dst=dst, degree=degree)

    def print(self) -> None:
        print("Sin")
        super().print()


class Cos(TrigonometryOperation):
    def __init__(self, dst: Argument, degree: Argument):
        super().__init__(dst=dst, degree=degree)

    def print(self) -> None:
        print("Cos")
        super().print()


class ExponentiationOperation(Instruction):
    def __init__(self, dst: Argument, base: Argument, num: Argument):
        self.dst = dst
        self.base = base
        self.num = num

    def print(self) -> None:
        print("ExponentiationOperation")
        self.dst.print()
        self.base.print()
        self.num.print()


class Logarithm(ExponentiationOperation):
    def __init__(self, dst: Argument, base: Argument, num: Argument):
        super().__init__(dst=dst, base=base, num=num)

    def print(self) -> None:
        print("Logarithm")
        super().print()


class Power(ExponentiationOperation):
    def __init__(self, dst: Argument, base: Argument, num: Argument):
        super().__init__(dst=dst, base=base, num=num)

    def print(self) -> None:
        print("Power")
        super().print()


class Modulo(Instruction):
    def __init__(self, dst: Argument, dividend: Argument, divisor: Argument):
        self.dst = dst
        self.dividend = dividend
        self.divisor = divisor

    def print(self) -> None:
        print("Modulo")
        self.dst.print()
        self.dividend.print()
        self.divisor.print()



from __future__ import annotations

from typing import List


class PythonCode:
    def __init__(self, indent_size: int):
        self.indent_size = indent_size
        self.indent: int = 0
        self.code_lines: List[str] = []

    def indent_left(self) -> None:
        self.indent -= self.indent_size

    def indent_right(self) -> None:
        self.indent += self.indent_size

    def add_line(self, line: str) -> None:
        self.code_lines.append(self.indent * " " + line + "\n")

    def add_newline(self) -> None:
        self.add_line("")

    def add_newlines(self, count: int) -> None:
        for _ in range(count):
            self.add_newline()

    def add_template(self, template: str, **kwargs: Any) -> None:
        lines = template.render(kwargs).splitlines()
        current_indent = 0
        for line in lines:
            space_count = len(line) - len(
                line.lstrip(" ")
            )  # templates should have 4 space indents per pep8
            indent_count = space_count // 4
            if indent_count < current_indent:
                for _ in range(current_indent - indent_count):
                    self.indent_left()
            elif indent_count > current_indent:
                for _ in range(indent_count - current_indent):
                    self.indent_right()
            self.add_line(line.strip())



from __future__ import annotations

from itertools import compress
from typing import TYPE_CHECKING, List

from aasm.generating.python_code import PythonCode
from aasm.intermediate.graph import (
    AgentConstantAmount,
    AgentPercentAmount,
    ConnectionConstantAmount,
    ConnectionDistExpAmount,
    ConnectionDistNormalAmount,
    ConnectionDistUniformAmount,
    MatrixGraph,
    StatisticalGraph,
)

if TYPE_CHECKING:
    from aasm.intermediate.graph import Graph


class PythonGraph(PythonCode):
    def __init__(self, indent_size: int, graph: Graph | None):
        super().__init__(indent_size)
        if graph:
            self.add_required_imports()
            self.add_newlines(2)
            self.generate_graph(graph)

    def add_required_imports(self) -> None:
        self.add_line("import random")
        self.add_line("import uuid")
        self.add_line("import numpy")

    def generate_graph(self, graph: Graph) -> None:
        match graph:
            case StatisticalGraph():
                self.add_statistical_graph(graph)

            case MatrixGraph():
                self.add_matrix_graph(graph)

            case _:
                raise Exception(f"Unknown graph type: {graph.print()}")

    def add_statistical_graph(self, graph: StatisticalGraph) -> None:
        self.add_line("def generate_graph_structure(domain):")
        self.indent_right()

        if not graph.agents:
            self.add_line("return []")
            self.indent_left()
            return

        num_agents_expr: List[str] = []
        for agent in graph.agents.values():
            match agent.amount:
                case AgentConstantAmount():
                    self.add_line(f"_num_{agent.name} = {agent.amount.value}")

                case AgentPercentAmount():
                    self.add_line(
                        f"_num_{agent.name} = round({agent.amount.value} / 100 * {graph.size})"
                    )

                case _:
                    raise Exception(
                        f"Unknown agent amount type: {agent.amount.print()}"
                    )

            num_agents_expr.append(f"_num_{agent.name}")

        self.add_line(f'num_agents = {" + ".join(num_agents_expr)}')
        self.add_line("random_id = str(uuid.uuid4())[:5]")
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]')
        self.add_line("agents = []")
        self.add_line("next_agent_idx = 0")
        for agent in graph.agents.values():
            self.add_line(f"for _ in range(_num_{agent.name}):")
            self.indent_right()

            match agent.connections:
                case ConnectionConstantAmount():
                    self.add_line(f"num_connections = {agent.connections.value}")

                case ConnectionDistNormalAmount():
                    self.add_line(
                        f"num_connections = int(numpy.random.normal({agent.connections.mean}, {agent.connections.std_dev}))"
                    )

                case ConnectionDistExpAmount():
                    self.add_line(
                        f"num_connections = int(numpy.random.exponential(1 / {agent.connections.lambda_}))"
                    )

                case ConnectionDistUniformAmount():
                    self.add_line(
                        f"num_connections = int(random.uniform({agent.connections.a}, {agent.connections.b}))"
                    )

                case _:
                    raise Exception(
                        f"Unknown connection amount: {agent.connections.print()}"
                    )

            self.add_line(
                "num_connections = max(min(num_connections, len(jids) - 1), 0)"
            )
            self.add_line("jid = jids[next_agent_idx]")
            self.add_line("agents.append({")
            self.indent_right()
            self.add_line('"jid": jid,')
            self.add_line(f'"type": "{agent.name}",')
            self.add_line(
                '"connections": random.sample([other_jid for other_jid in jids if other_jid != jid], num_connections),'
            )
            self.indent_left()
            self.add_line("})")
            self.add_line("next_agent_idx += 1")
            self.indent_left()

        self.add_line("return agents")
        self.indent_left()

    def add_matrix_graph(self, graph: MatrixGraph):
        self.add_line("def generate_graph_structure(domain):")
        self.indent_right()
        if not graph.agents:
            self.add_line("return []")
            self.indent_left()
            return
        if graph.is_scale_defined():
            self.add_line(f"scale_factor = {graph.scale}")
        else:
            self.add_line(f"scale_factor = 1")
        self.add_line(f"n_agent_types = {len(graph.agents)}")
        self.add_line("graph_size = scale_factor * n_agent_types")
        self.add_line("random_id = str(uuid.uuid4())[:5]")
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(graph_size)]')
        self.add_line("agents = []")
        self.add_line("indx_sets = []")
        for agent in graph.agents:
            adj_indx = [True if x == 1 else False for x in agent.adj_row.row]
            adj_indx = list(compress(range(len(adj_indx)), adj_indx))
            self.add_line(f"indx_sets.append({adj_indx})")

        self.add_line("for base_agent_index in range(n_agent_types):")
        self.indent_right()
        self.add_line("indices = indx_sets[base_agent_index]")
        self.add_line("for shift in range(1, scale_factor):")
        self.indent_right()
        self.add_line(
            f"indices.append((base_agent_index + shift * n_agent_types) % graph_size)"
        )
        self.indent_left()
        self.add_line("for shift in range(scale_factor):")
        self.indent_right()
        self.add_line("jid = jids[base_agent_index + shift * n_agent_types]")
        self.add_line("connections = []")
        self.add_line("for i in indices:")
        self.indent_right()
        self.add_line(
            f"connections.append(jids[(i + shift * n_agent_types) % graph_size])"
        )
        self.indent_left()
        self.add_line("agents.append({")
        self.indent_right()
        self.add_line('"jid": jid,')
        self.add_line(f'"type": "{agent.name}",')
        self.add_line('"connections": connections,')
        self.indent_left()
        self.add_line("})")
        self.indent_left()

        self.indent_left()
        self.add_line("return agents")
        self.indent_left()



from __future__ import annotations

from typing import Any, Dict


class AgentAmount:
    def __init__(self, value: str):
        self.value = value

    def print(self) -> None:
        print(f"AgentAmount value = {self.value}")


class AgentConstantAmount(AgentAmount):
    def __init__(self, value: str):
        super().__init__(value)

    def print(self) -> None:
        print("AgentConstantAmount")
        super().print()


class AgentPercentAmount(AgentAmount):
    def __init__(self, value: str):
        super().__init__(value)

    def print(self) -> None:
        print("AgentPercentAmount")
        super().print()


class ConnectionAmount:
    def print(self) -> None:
        raise NotImplementedError()


class ConnectionConstantAmount(ConnectionAmount):
    def __init__(self, value: str):
        self.value = value

    def print(self) -> None:
        print(f"ConnectionConstantAmount value = {self.value}")


class ConnectionDistNormalAmount(ConnectionAmount):
    def __init__(self, mean: str, std_dev: str):
        self.mean = mean
        self.std_dev = std_dev

    def print(self) -> None:
        print(
            f"ConnectionDistNormalAmount mean = {self.mean}, std_dev = {self.std_dev}"
        )


class ConnectionDistExpAmount(ConnectionAmount):
    def __init__(self, lambda_: str):
        self.lambda_ = lambda_

    def print(self) -> None:
        print(f"ConnectionDistExpAmount lambda = {self.lambda_}")


class ConnectionDistUniformAmount(ConnectionAmount):
    def __init__(self, a: str, b: str):
        self.a = a
        self.b = b

    def print(self) -> None:
        print(f"ConnectionDistUniformAmount a = {self.a}, b = {self.b}")


class AdjRow:
    def __init__(self, row: List[int]):
        self.row = row

    def print(self) -> None:
        print("AdjRow")
        for key, value in self.row.items():
            print(f"{key} = {value}")


class MatrixAgent:
    def __init__(self, name: str, adj_row: AdjRow):
        self.name = name
        self.adj_row = adj_row

    def print(self) -> None:
        print(f"MatrixAgent name = {self.name}")
        self.adj_row.print()


class StatisticalAgent:
    def __init__(self, name: str, amount: AgentAmount, connections: ConnectionAmount):
        self.name = name
        self.amount = amount
        self.connections = connections

    def print(self) -> None:
        print(f"StatisticalAgent name = {self.name}")
        self.amount.print()
        self.connections.print()


class Graph:
    def __init__(self):
        self.size = None
        self.scale = None

    def set_size(self, size: int) -> None:
        self.size = size

    def is_size_defined(self) -> bool:
        return self.size is not None

    def set_scale(self, scale: int) -> None:
        self.scale = scale

    def is_scale_defined(self) -> bool:
        return self.scale is not None

    def add_agent(self, graph_agent: Any) -> None:
        raise NotImplementedError()

    def print(self) -> None:
        print(f"Graph size = {self.size}")


class StatisticalGraph(Graph):
    def __init__(self):
        super().__init__()
        self.agents: Dict[str, StatisticalAgent] = {}

    def add_agent(self, graph_agent: StatisticalAgent) -> None:
        self.agents[graph_agent.name] = graph_agent

    def is_agent_defined(self, agent_type: str) -> bool:
        return agent_type in self.agents

    def is_agent_percent_amount_used(self) -> bool:
        for agent in self.agents.values():
            if isinstance(agent.amount, AgentPercentAmount):
                return True
        return False

    def print(self) -> None:
        super().print()
        print("StatisticalGraph")
        for agent in self.agents.values():
            agent.print()


class MatrixGraph(Graph):
    def __init__(self):
        super().__init__()
        self.agents = []

    def add_agent(self, graph_agent: MatrixAgent) -> None:
        self.agents.append(graph_agent)

    def is_agent_defined(self, agent_type: str) -> bool:
        return any(self.agents, lambda agent: agent.name == agent_type)

    def set_scale(self, scale: int) -> None:
        self.scale = scale

    def is_scale_defined(self) -> bool:
        return self.scale is not None

    def print(self) -> None:
        super().print()
        print("MatrixGraph")
        for agent in self.agents.values():
            agent.print()



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Subset

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_SUBS(state: State, arg1: str, arg2: str, arg3: str) -> None:
    state.require(
        state.in_action, "Not inside any action.", f"SUBS can be used inside actions."
    )
    dst_list = Argument(state, arg1)
    src_list = Argument(state, arg2)
    num = Argument(state, arg3)
    state.require(
        dst_list.list_subset_context(src_list, num),
        "Mismatched types in the subset context.",
        f"ARG1 {dst_list.explain()}, ARG2 {src_list.explain()}, ARG3 {num.explain()}",
    )

    state.last_action.add_instruction(Subset(dst_list, src_list, num))



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Modulo

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_MOD(state: State, arg1: str, arg2: str, arg3: str) -> None:
    state.require(
        state.in_action, "Not inside any action", f"MOD can be used inside actions."
    )
    dst = Argument(state, arg1)
    dividend = Argument(state, arg2)
    divisor = Argument(state, arg3)
    state.require(
        dst.math_modulo_context(dividend, divisor),
        "Mismatched types in the math modulo statement.",
        f"ARG1 {dst.explain()}, ARG2 {dividend.explain()}, ARG3 {divisor.explain()}",
    )

    state.last_action.add_instruction(Modulo(dst, dividend, divisor))



import copy
import datetime
import random
import httpx
import numpy
import orjson
import spade
import sys


class grass(spade.agent.Agent):
    def __init__(self, jid, password, backup_url = None, backup_period = 60, backup_delay = 0, logger = None, **kwargs):
        super().__init__(jid, password, verify_security=False)
        if logger: logger.debug(f'[{jid}] Received parameters: jid: {jid}, password: {password}, backup_url: {backup_url}, backup_period: {backup_period}, backup_delay: {backup_delay}, kwargs: {kwargs}')
        self.logger = logger
        self.backup_url = backup_url
        self.backup_period = backup_period
        self.backup_delay = backup_delay
        self.connections = kwargs.get("connections", [])
        self.msgRCount = self.limit_number(kwargs.get("msgRCount", 0))
        self.msgSCount = self.limit_number(kwargs.get("msgSCount", 0))
        self.dead_sheep = kwargs.get("dead_sheep", [])
        self.alive_sheep = kwargs.get("alive_sheep", [])
        self.dead_wolf = kwargs.get("dead_wolf", [])
        self.alive_wolf = kwargs.get("alive_wolf", [])
        self.spawn_candidate = kwargs.get("spawn_candidate", [])
        self.field = kwargs.get("field", [])
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after initialization: {self.__dict__}')
    
    @property
    def connCount(self):
        return self.limit_number(len(self.connections))
    
    def limit_number(self, value):
        return float(max(-2147483648, min(value, 2147483647)))
    
    def get_json_from_spade_message(self, msg):
        return orjson.loads(msg.body)
    
    def get_spade_message(self, receiver_jid, body):
        msg = spade.message.Message(to=receiver_jid)
        body["sender"] = str(self.jid)
        msg.metadata["type"] = body["type"]
        msg.metadata["performative"] = body["performative"]
        msg.body = str(orjson.dumps(body), encoding="utf-8")
        return msg
    
    def setup(self):
        if self.backup_url:
            BackupBehaviour_template = spade.template.Template()
            BackupBehaviour_template.set_metadata("reserved", "no_message_match")
            self.add_behaviour(self.BackupBehaviour(start_at=datetime.datetime.now() + datetime.timedelta(seconds=self.backup_delay), period=self.backup_period), BackupBehaviour_template)
        setup_template = spade.template.Template()
        setup_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.setup(), setup_template)
        regrow_grass_template = spade.template.Template()
        regrow_grass_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.regrow_grass(period=10), regrow_grass_template)
        recv_spawn_template = spade.template.Template()
        recv_spawn_template.set_metadata("type", "spawn")
        recv_spawn_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_spawn(), recv_spawn_template)
        recv_broadcast_template = spade.template.Template()
        recv_broadcast_template.set_metadata("type", "broadcast")
        recv_broadcast_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_broadcast(), recv_broadcast_template)
        eat_grass_template = spade.template.Template()
        eat_grass_template.set_metadata("type", "eat")
        eat_grass_template.set_metadata("performative", "Request")
        self.add_behaviour(self.eat_grass(), eat_grass_template)
        recv_kill_template = spade.template.Template()
        recv_kill_template.set_metadata("type", "kill")
        recv_kill_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_kill(), recv_kill_template)
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after setup: {self.__dict__}')
    
    class BackupBehaviour(spade.behaviour.PeriodicBehaviour):
        def __init__(self, start_at, period):
            super().__init__(start_at=start_at, period=period)
            self.http_client = httpx.AsyncClient(timeout=period)
        
        async def run(self):
            data = {
                "__timestamp__": int(datetime.datetime.timestamp(datetime.datetime.utcnow())),
                "jid": str(self.agent.jid),
                "type": "grass",
                "floats": {
                    "msgRCount": self.agent.msgRCount,
                    "msgSCount": self.agent.msgSCount,
                    "connCount": self.agent.connCount,
                },
                "enums": {
                },
                "connections": {
                    "connections": self.agent.connections,
                    "dead_sheep": self.agent.dead_sheep,
                    "alive_sheep": self.agent.alive_sheep,
                    "dead_wolf": self.agent.dead_wolf,
                    "alive_wolf": self.agent.alive_wolf,
                    "spawn_candidate": self.agent.spawn_candidate,
                },
                "messages": {
                },
                "float_lists": {
                    "field": self.agent.field,
                },
            }
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Sending backup data: {data}')
            try:
                await self.http_client.post(self.agent.backup_url, headers={"Content-Type": "application/json"}, data=orjson.dumps(data))
            except Exception as e:
                if self.agent.logger: self.agent.logger.error(f'[{self.agent.jid}] Backup error type: {e.__class__}, additional info: {e}')
    
    class setup(spade.behaviour.OneShotBehaviour):
        def setup_field(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action setup_field')
            
            # float declaration
            it = self.agent.limit_number(0)
            
            # float declaration
            amt = self.agent.limit_number(0)
            
            # while less than
            while self.agent.limit_number(it) < self.agent.limit_number(10000):
                
                # uniform distribution
                amt = self.agent.limit_number(random.uniform(self.agent.limit_number(0), self.agent.limit_number(10)))
                
                # add element
                self.agent.field.append(amt)
                
                # add
                it = self.agent.limit_number(it + self.agent.limit_number(1))
        
        async def broadcast(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action broadcast')
            send = { "type": "broadcast", "performative": "Inform", "typ": 0.0, "state": 0.0, }
            
            # set
            send["typ"] = self.agent.limit_number(0)
            
            # send
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.connections}')
            for receiver in self.agent.connections:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            self.setup_field()
            await self.broadcast()
    
    class regrow_grass(spade.behaviour.PeriodicBehaviour):
        def regrow(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action regrow')
            
            # float declaration
            it = self.agent.limit_number(0)
            
            # float declaration
            amt = self.agent.limit_number(0)
            
            # while less than
            while self.agent.limit_number(it) < self.agent.limit_number(10000):
                
                # list read
                if int(self.agent.limit_number(round(self.agent.limit_number(it)))) < 0 or int(self.agent.limit_number(round(self.agent.limit_number(it)))) >= int(self.agent.limit_number(len(self.agent.field))):
                    if self.agent.logger: self.agent.logger.warning(f'[{self.agent.jid}] Incorrect index (rounded, either negative or bigger than the list size): {int(self.agent.limit_number(round(self.agent.limit_number(it))))}')
                    return
                amt = self.agent.limit_number(self.agent.field[int(self.agent.limit_number(round(self.agent.limit_number(it))))])
                
                # add
                amt = self.agent.limit_number(amt + self.agent.limit_number(0.1))
                
                # if greater than
                if self.agent.limit_number(amt) > self.agent.limit_number(10):
                    
                    # set
                    amt = self.agent.limit_number(10)
                
                # list write
                if int(self.agent.limit_number(round(self.agent.limit_number(it)))) < 0 or int(self.agent.limit_number(round(self.agent.limit_number(it)))) >= int(self.agent.limit_number(len(self.agent.field))):
                    if self.agent.logger: self.agent.logger.warning(f'[{self.agent.jid}] Incorrect index (rounded, either negative or bigger than the list size): {int(self.agent.limit_number(round(self.agent.limit_number(it))))}')
                    return
                self.agent.field[int(self.agent.limit_number(round(self.agent.limit_number(it))))] = self.agent.limit_number(amt)
                
                # add
                it = self.agent.limit_number(it + self.agent.limit_number(1))
        
        async def run(self):
            self.regrow()
    
    class recv_spawn(spade.behaviour.CyclicBehaviour):
        def choose_spawn(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action choose_spawn')
            
            # clear
            self.agent.spawn_candidate.clear()
            
            # if equal
            if self.agent.limit_number(rcv["typ"]) == self.agent.limit_number(1):
                
                # subset
                if int(self.agent.limit_number(round(self.agent.limit_number(1)))) <= 0:
                    if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Non-positive subset size (rounded): {int(self.agent.limit_number(round(self.agent.limit_number(1))))}')
                    return
                self.agent.spawn_candidate = [copy.deepcopy(elem) for elem in random.sample(self.agent.dead_sheep, min(int(self.agent.limit_number(round(self.agent.limit_number(1)))), int(self.agent.limit_number(len(self.agent.dead_sheep)))))]
                
                # remove element
                if "spawn_candidate" in self.agent.dead_sheep: self.agent.dead_sheep.remove("spawn_candidate")
                
                # add element
                self.agent.alive_sheep.append("spawn_candidate")
            
            # if equal
            if self.agent.limit_number(rcv["typ"]) == self.agent.limit_number(2):
                
                # subset
                if int(self.agent.limit_number(round(self.agent.limit_number(1)))) <= 0:
                    if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Non-positive subset size (rounded): {int(self.agent.limit_number(round(self.agent.limit_number(1))))}')
                    return
                self.agent.spawn_candidate = [copy.deepcopy(elem) for elem in random.sample(self.agent.dead_wolf, min(int(self.agent.limit_number(round(self.agent.limit_number(1)))), int(self.agent.limit_number(len(self.agent.dead_wolf)))))]
                
                # remove element
                if "spawn_candidate" in self.agent.dead_wolf: self.agent.dead_wolf.remove("spawn_candidate")
                
                # add element
                self.agent.alive_wolf.append("spawn_candidate")
        
        async def spawn(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action spawn')
            send = { "type": "spawn", "performative": "Inform", "typ": 0.0, }
            
            # send
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.spawn_candidate}')
            for receiver in self.agent.spawn_candidate:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.choose_spawn(rcv)
                await self.spawn(rcv)
    
    class recv_broadcast(spade.behaviour.CyclicBehaviour):
        def add_agent(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action add_agent')
            
            # if equal
            if self.agent.limit_number(rcv["typ"]) == self.agent.limit_number(1):
                
                # if equal
                if self.agent.limit_number(rcv["state"]) == self.agent.limit_number(0):
                    
                    # remove element
                    if "RCV.jid" in self.agent.alive_sheep: self.agent.alive_sheep.remove("RCV.jid")
                    
                    # add element
                    self.agent.dead_sheep.append("RCV.jid")
                
                # if equal
                if self.agent.limit_number(rcv["state"]) == self.agent.limit_number(1):
                    
                    # remove element
                    if "RCV.jid" in self.agent.dead_sheep: self.agent.dead_sheep.remove("RCV.jid")
                    
                    # add element
                    self.agent.alive_sheep.append("RCV.jid")
            
            # if equal
            if self.agent.limit_number(rcv["typ"]) == self.agent.limit_number(2):
                
                # if equal
                if self.agent.limit_number(rcv["state"]) == self.agent.limit_number(0):
                    
                    # remove element
                    if "RCV.jid" in self.agent.alive_wolf: self.agent.alive_wolf.remove("RCV.jid")
                    
                    # add element
                    self.agent.dead_wolf.append("RCV.jid")
                
                # if equal
                if self.agent.limit_number(rcv["state"]) == self.agent.limit_number(1):
                    
                    # remove element
                    if "RCV.jid" in self.agent.dead_wolf: self.agent.dead_wolf.remove("RCV.jid")
                    
                    # add element
                    self.agent.alive_wolf.append("RCV.jid")
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.add_agent(rcv)
    
    class eat_grass(spade.behaviour.CyclicBehaviour):
        async def parse_eat(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action parse_eat')
            send = { "type": "eat", "performative": "Inform", "energy": 0.0, }
            
            # float declaration
            index = self.agent.limit_number(rcv["x"])
            
            # multiply
            index = self.agent.limit_number(index * self.agent.limit_number(99))
            
            # add
            index = self.agent.limit_number(index + self.agent.limit_number(rcv["y"]))
            
            # float declaration
            grass_val = self.agent.limit_number(0)
            
            # list read
            if int(self.agent.limit_number(round(self.agent.limit_number(index)))) < 0 or int(self.agent.limit_number(round(self.agent.limit_number(index)))) >= int(self.agent.limit_number(len(self.agent.field))):
                if self.agent.logger: self.agent.logger.warning(f'[{self.agent.jid}] Incorrect index (rounded, either negative or bigger than the list size): {int(self.agent.limit_number(round(self.agent.limit_number(index))))}')
                return
            grass_val = self.agent.limit_number(self.agent.field[int(self.agent.limit_number(round(self.agent.limit_number(index))))])
            
            # float declaration
            send_flag = self.agent.limit_number(0)
            
            # if greater than
            if self.agent.limit_number(grass_val) > self.agent.limit_number(1):
                
                # set
                send_flag = self.agent.limit_number(1)
                
                # set
                send["energy"] = self.agent.limit_number(1)
                
                # subtract
                grass_val = self.agent.limit_number(grass_val - self.agent.limit_number(1))
                
                # list write
                if int(self.agent.limit_number(round(self.agent.limit_number(index)))) < 0 or int(self.agent.limit_number(round(self.agent.limit_number(index)))) >= int(self.agent.limit_number(len(self.agent.field))):
                    if self.agent.logger: self.agent.logger.warning(f'[{self.agent.jid}] Incorrect index (rounded, either negative or bigger than the list size): {int(self.agent.limit_number(round(self.agent.limit_number(index))))}')
                    return
                self.agent.field[int(self.agent.limit_number(round(self.agent.limit_number(index))))] = self.agent.limit_number(grass_val)
            
            # if equal
            if self.agent.limit_number(send_flag) == self.agent.limit_number(0):
                
                # set
                send["energy"] = self.agent.limit_number(0)
            
            # send
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {"RCV.jid"}')
            await self.send(self.agent.get_spade_message("RCV.jid", send))
            self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                await self.parse_eat(rcv)
    
    class recv_kill(spade.behaviour.CyclicBehaviour):
        def kill_creature(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action kill_creature')
            
            # if in list
            if "RCV.jid" in self.agent.alive_sheep:
                
                # remove element
                if "RCV.jid" in self.agent.alive_sheep: self.agent.alive_sheep.remove("RCV.jid")
                
                # add element
                self.agent.dead_sheep.append("RCV.jid")
            
            # if in list
            if "RCV.jid" in self.agent.alive_wolf:
                
                # remove element
                if "RCV.jid" in self.agent.alive_wolf: self.agent.alive_wolf.remove("RCV.jid")
                
                # add element
                self.agent.dead_wolf.append("RCV.jid")
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.kill_creature(rcv)
    


class sheep(spade.agent.Agent):
    def __init__(self, jid, password, backup_url = None, backup_period = 60, backup_delay = 0, logger = None, **kwargs):
        super().__init__(jid, password, verify_security=False)
        if logger: logger.debug(f'[{jid}] Received parameters: jid: {jid}, password: {password}, backup_url: {backup_url}, backup_period: {backup_period}, backup_delay: {backup_delay}, kwargs: {kwargs}')
        self.logger = logger
        self.backup_url = backup_url
        self.backup_period = backup_period
        self.backup_delay = backup_delay
        self.connections = kwargs.get("connections", [])
        self.msgRCount = self.limit_number(kwargs.get("msgRCount", 0))
        self.msgSCount = self.limit_number(kwargs.get("msgSCount", 0))
        self.energy = self.limit_number(kwargs.get("energy", 100))
        self.x = self.limit_number(kwargs.get("x", random.uniform(self.limit_number(0), self.limit_number(99))))
        self.y = self.limit_number(kwargs.get("y", random.uniform(self.limit_number(0), self.limit_number(99))))
        self.heading = self.limit_number(kwargs.get("heading", random.uniform(self.limit_number(0), self.limit_number(180))))
        self.invert = kwargs.get("invert", random.choices(["YES", "NO"], [0, 100])[0])
        self.dead = kwargs.get("dead", random.choices(["YES", "NO"], [0, 100])[0])
        self.field = kwargs.get("field", [])
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after initialization: {self.__dict__}')
    
    @property
    def connCount(self):
        return self.limit_number(len(self.connections))
    
    def limit_number(self, value):
        return float(max(-2147483648, min(value, 2147483647)))
    
    def get_json_from_spade_message(self, msg):
        return orjson.loads(msg.body)
    
    def get_spade_message(self, receiver_jid, body):
        msg = spade.message.Message(to=receiver_jid)
        body["sender"] = str(self.jid)
        msg.metadata["type"] = body["type"]
        msg.metadata["performative"] = body["performative"]
        msg.body = str(orjson.dumps(body), encoding="utf-8")
        return msg
    
    def setup(self):
        if self.backup_url:
            BackupBehaviour_template = spade.template.Template()
            BackupBehaviour_template.set_metadata("reserved", "no_message_match")
            self.add_behaviour(self.BackupBehaviour(start_at=datetime.datetime.now() + datetime.timedelta(seconds=self.backup_delay), period=self.backup_period), BackupBehaviour_template)
        setup_template = spade.template.Template()
        setup_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.setup(), setup_template)
        move_around_template = spade.template.Template()
        move_around_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.move_around(period=10), move_around_template)
        reproduce_template = spade.template.Template()
        reproduce_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.reproduce(period=20), reproduce_template)
        recv_spawn_template = spade.template.Template()
        recv_spawn_template.set_metadata("type", "spawn")
        recv_spawn_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_spawn(), recv_spawn_template)
        recv_kill_template = spade.template.Template()
        recv_kill_template.set_metadata("type", "kill")
        recv_kill_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_kill(), recv_kill_template)
        grass_eat_template = spade.template.Template()
        grass_eat_template.set_metadata("type", "eat")
        grass_eat_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.grass_eat(), grass_eat_template)
        recv_broadcast_template = spade.template.Template()
        recv_broadcast_template.set_metadata("type", "broadcast")
        recv_broadcast_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_broadcast(), recv_broadcast_template)
        get_eaten_request_template = spade.template.Template()
        get_eaten_request_template.set_metadata("type", "eat")
        get_eaten_request_template.set_metadata("performative", "Request")
        self.add_behaviour(self.get_eaten_request(), get_eaten_request_template)
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after setup: {self.__dict__}')
    
    class BackupBehaviour(spade.behaviour.PeriodicBehaviour):
        def __init__(self, start_at, period):
            super().__init__(start_at=start_at, period=period)
            self.http_client = httpx.AsyncClient(timeout=period)
        
        async def run(self):
            data = {
                "__timestamp__": int(datetime.datetime.timestamp(datetime.datetime.utcnow())),
                "jid": str(self.agent.jid),
                "type": "sheep",
                "floats": {
                    "msgRCount": self.agent.msgRCount,
                    "msgSCount": self.agent.msgSCount,
                    "connCount": self.agent.connCount,
                    "energy": self.agent.energy,
                    "x": self.agent.x,
                    "y": self.agent.y,
                    "heading": self.agent.heading,
                },
                "enums": {
                    "invert": self.agent.invert,
                    "dead": self.agent.dead,
                },
                "connections": {
                    "connections": self.agent.connections,
                    "field": self.agent.field,
                },
                "messages": {
                },
                "float_lists": {
                },
            }
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Sending backup data: {data}')
            try:
                await self.http_client.post(self.agent.backup_url, headers={"Content-Type": "application/json"}, data=orjson.dumps(data))
            except Exception as e:
                if self.agent.logger: self.agent.logger.error(f'[{self.agent.jid}] Backup error type: {e.__class__}, additional info: {e}')
    
    class setup(spade.behaviour.OneShotBehaviour):
        async def broadcast(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action broadcast')
            send = { "type": "broadcast", "performative": "Inform", "typ": 0.0, "state": 0.0, }
            
            # set
            send["typ"] = self.agent.limit_number(1)
            
            # if equal
            if self.agent.dead == "YES":
                
                # set
                send["state"] = self.agent.limit_number(0)
            
            # if equal
            if self.agent.dead == "NO":
                
                # set
                send["state"] = self.agent.limit_number(1)
            
            # send
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.connections}')
            for receiver in self.agent.connections:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            await self.broadcast()
    
    class move_around(spade.behaviour.PeriodicBehaviour):
        def wiggle(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action wiggle')
            
            # if equal
            if self.agent.dead == "NO":
                
                # float declaration
                wiglR = self.agent.limit_number(0)
                
                # float declaration
                wiglL = self.agent.limit_number(0)
                
                # uniform distribution
                wiglR = self.agent.limit_number(random.uniform(self.agent.limit_number(0), self.agent.limit_number(90)))
                
                # uniform distribution
                wiglL = self.agent.limit_number(random.uniform(self.agent.limit_number(0), self.agent.limit_number(90)))
                
                # float declaration
                change_flag = self.agent.limit_number(0)
                
                # if equal
                if self.agent.invert == "YES":
                    
                    # set
                    change_flag = self.agent.limit_number(1)
                    
                    # add
                    self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(wiglL))
                    
                    # subtract
                    self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(wiglR))
                    
                    # if less than
                    if self.agent.limit_number(self.agent.heading) < self.agent.limit_number(0):
                        
                        # set
                        self.agent.invert = "NO"
                        
                        # add
                        self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(180))
                    
                    # if greater than
                    if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(180):
                        
                        # set
                        self.agent.invert = "NO"
                        
                        # subtract
                        self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(180))
                
                # if equal
                if self.agent.invert == "NO":
                    
                    # if equal
                    if self.agent.limit_number(change_flag) == self.agent.limit_number(0):
                        
                        # add
                        self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(wiglL))
                        
                        # subtract
                        self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(wiglR))
                        
                        # if less than
                        if self.agent.limit_number(self.agent.heading) < self.agent.limit_number(0):
                            
                            # set
                            self.agent.invert = "YES"
                            
                            # add
                            self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(180))
                        
                        # if greater than
                        if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(180):
                            
                            # set
                            self.agent.invert = "YES"
                            
                            # subtract
                            self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(180))
        
        def move(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action move')
            
            # if equal
            if self.agent.dead == "NO":
                
                # float declaration
                a = self.agent.limit_number(0)
                
                # float declaration
                b = self.agent.limit_number(0)
                
                # sin
                a = self.agent.limit_number(numpy.sin(numpy.deg2rad(self.agent.limit_number(self.agent.heading))))
                
                # sin
                b = self.agent.limit_number(numpy.sin(numpy.deg2rad(self.agent.limit_number(self.agent.heading))))
                
                # float declaration
                newx = self.agent.limit_number(self.agent.x)
                
                # float declaration
                newy = self.agent.limit_number(self.agent.y)
                
                # if greater than
                if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(90):
                    
                    # subtract
                    newx = self.agent.limit_number(newx - self.agent.limit_number(a))
                
                # if greater than or equal
                if self.agent.limit_number(self.agent.heading) <= self.agent.limit_number(90):
                    
                    # add
                    newx = self.agent.limit_number(newx + self.agent.limit_number(a))
                
                # if equal
                if self.agent.invert == "YES":
                    
                    # subtract
                    newy = self.agent.limit_number(newy - self.agent.limit_number(b))
                
                # if equal
                if self.agent.invert == "NO":
                    
                    # add
                    newy = self.agent.limit_number(newy + self.agent.limit_number(b))
                
                # float declaration
                bump = self.agent.limit_number(0)
                
                # if less than
                if self.agent.limit_number(newx) < self.agent.limit_number(0):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if less than
                if self.agent.limit_number(newy) < self.agent.limit_number(0):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if greater than
                if self.agent.limit_number(newx) > self.agent.limit_number(99):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if greater than
                if self.agent.limit_number(newy) > self.agent.limit_number(99):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if equal
                if self.agent.limit_number(bump) == self.agent.limit_number(0):
                    
                    # set
                    self.agent.x = self.agent.limit_number(newx)
                    
                    # set
                    self.agent.y = self.agent.limit_number(newy)
                    
                    # subtract
                    self.agent.energy = self.agent.limit_number(self.agent.energy - self.agent.limit_number(1))
                
                # if equal
                if self.agent.limit_number(bump) == self.agent.limit_number(1):
                    
                    # add
                    self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(90))
                    
                    # if greater than
                    if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(180):
                        
                        # subtract
                        self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(180))
                        
                        # float declaration
                        flag = self.agent.limit_number(0)
                        
                        # if equal
                        if self.agent.invert == "YES":
                            
                            # set
                            flag = self.agent.limit_number(1)
                            
                            # set
                            self.agent.invert = "NO"
                        
                        # if equal
                        if self.agent.invert == "NO":
                            
                            # if equal
                            if self.agent.limit_number(flag) == self.agent.limit_number(0):
                                
                                # set
                                self.agent.invert = "YES"
        
        def check_if_dead(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action check_if_dead')
            
            # if greater than or equal
            if self.agent.limit_number(self.agent.energy) <= self.agent.limit_number(0):
                
                # set
                self.agent.dead = "YES"
        
        async def eat(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action eat')
            send = { "type": "eat", "performative": "Request", "x": 0.0, "y": 0.0, }
            
            # if equal
            if self.agent.dead == "NO":
                
                # set
                send["x"] = self.agent.limit_number(self.agent.x)
                
                # set
                send["y"] = self.agent.limit_number(self.agent.y)
                
                # send
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.field}')
                for receiver in self.agent.field:
                    await self.send(self.agent.get_spade_message(receiver, send))
                    self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            self.wiggle()
            self.move()
            self.check_if_dead()
            await self.eat()
    
    class reproduce(spade.behaviour.PeriodicBehaviour):
        async def hatch(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action hatch')
            send = { "type": "spawn", "performative": "Inform", "typ": 0.0, }
            
            # if equal
            if self.agent.dead == "NO":
                
                # if equal
                if self.agent.limit_number(self.agent.energy) == self.agent.limit_number(100):
                    
                    # set
                    send["typ"] = self.agent.limit_number(1)
                    
                    # send
                    if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.field}')
                    for receiver in self.agent.field:
                        await self.send(self.agent.get_spade_message(receiver, send))
                        self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            await self.hatch()
    
    class recv_spawn(spade.behaviour.CyclicBehaviour):
        def spawn(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action spawn')
            
            # set
            self.agent.dead = "NO"
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.spawn(rcv)
    
    class recv_kill(spade.behaviour.CyclicBehaviour):
        async def die(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action die')
            send = { "type": "kill", "performative": "Inform", }
            
            # set
            self.agent.dead = "YES"
            
            # send
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.field}')
            for receiver in self.agent.field:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                await self.die(rcv)
    
    class grass_eat(spade.behaviour.CyclicBehaviour):
        def change_energy(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action change_energy')
            
            # add
            self.agent.energy = self.agent.limit_number(self.agent.energy + self.agent.limit_number(rcv["energy"]))
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.change_energy(rcv)
    
    class recv_broadcast(spade.behaviour.CyclicBehaviour):
        def xd(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action xd')
            
            # if equal
            if self.agent.limit_number(rcv["typ"]) == self.agent.limit_number(0):
                
                # add element
                self.agent.field.append("RCV.jid")
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.xd(rcv)
    
    class get_eaten_request(spade.behaviour.CyclicBehaviour):
        async def kappa(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action kappa')
            send = { "type": "eat", "performative": "Inform", "energy": 0.0, }
            
            # if equal
            if self.agent.dead == "NO":
                
                # float declaration
                lowx = self.agent.limit_number(self.agent.x)
                
                # float declaration
                lowy = self.agent.limit_number(self.agent.y)
                
                # float declaration
                highx = self.agent.limit_number(self.agent.x)
                
                # float declaration
                highy = self.agent.limit_number(self.agent.y)
                
                # subtract
                lowx = self.agent.limit_number(lowx - self.agent.limit_number(1))
                
                # subtract
                lowy = self.agent.limit_number(lowy - self.agent.limit_number(1))
                
                # add
                highx = self.agent.limit_number(highx + self.agent.limit_number(1))
                
                # add
                highy = self.agent.limit_number(highy + self.agent.limit_number(1))
                
                # if less than
                if self.agent.limit_number(highx) < self.agent.limit_number(rcv["x"]):
                    
                    # if less than
                    if self.agent.limit_number(highy) < self.agent.limit_number(rcv["y"]):
                        
                        # if greater than
                        if self.agent.limit_number(lowx) > self.agent.limit_number(rcv["x"]):
                            
                            # if greater than
                            if self.agent.limit_number(lowy) > self.agent.limit_number(rcv["y"]):
                                
                                # set
                                send["energy"] = self.agent.limit_number(1)
                                
                                # send
                                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {"RCV.jid"}')
                                await self.send(self.agent.get_spade_message("RCV.jid", send))
                                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                await self.kappa(rcv)
    


class wolf(spade.agent.Agent):
    def __init__(self, jid, password, backup_url = None, backup_period = 60, backup_delay = 0, logger = None, **kwargs):
        super().__init__(jid, password, verify_security=False)
        if logger: logger.debug(f'[{jid}] Received parameters: jid: {jid}, password: {password}, backup_url: {backup_url}, backup_period: {backup_period}, backup_delay: {backup_delay}, kwargs: {kwargs}')
        self.logger = logger
        self.backup_url = backup_url
        self.backup_period = backup_period
        self.backup_delay = backup_delay
        self.connections = kwargs.get("connections", [])
        self.msgRCount = self.limit_number(kwargs.get("msgRCount", 0))
        self.msgSCount = self.limit_number(kwargs.get("msgSCount", 0))
        self.energy = self.limit_number(kwargs.get("energy", 100))
        self.x = self.limit_number(kwargs.get("x", random.uniform(self.limit_number(0), self.limit_number(99))))
        self.y = self.limit_number(kwargs.get("y", random.uniform(self.limit_number(0), self.limit_number(99))))
        self.heading = self.limit_number(kwargs.get("heading", random.uniform(self.limit_number(0), self.limit_number(180))))
        self.invert = kwargs.get("invert", random.choices(["YES", "NO"], [0, 100])[0])
        self.dead = kwargs.get("dead", random.choices(["YES", "NO"], [0, 100])[0])
        self.is_eating = kwargs.get("is_eating", random.choices(["YES", "NO"], [0, 100])[0])
        self.field = kwargs.get("field", [])
        self.sheep = kwargs.get("sheep", [])
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after initialization: {self.__dict__}')
    
    @property
    def connCount(self):
        return self.limit_number(len(self.connections))
    
    def limit_number(self, value):
        return float(max(-2147483648, min(value, 2147483647)))
    
    def get_json_from_spade_message(self, msg):
        return orjson.loads(msg.body)
    
    def get_spade_message(self, receiver_jid, body):
        msg = spade.message.Message(to=receiver_jid)
        body["sender"] = str(self.jid)
        msg.metadata["type"] = body["type"]
        msg.metadata["performative"] = body["performative"]
        msg.body = str(orjson.dumps(body), encoding="utf-8")
        return msg
    
    def setup(self):
        if self.backup_url:
            BackupBehaviour_template = spade.template.Template()
            BackupBehaviour_template.set_metadata("reserved", "no_message_match")
            self.add_behaviour(self.BackupBehaviour(start_at=datetime.datetime.now() + datetime.timedelta(seconds=self.backup_delay), period=self.backup_period), BackupBehaviour_template)
        setup_template = spade.template.Template()
        setup_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.setup(), setup_template)
        move_around_template = spade.template.Template()
        move_around_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.move_around(period=10), move_around_template)
        recv_spawn_template = spade.template.Template()
        recv_spawn_template.set_metadata("type", "spawn")
        recv_spawn_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_spawn(), recv_spawn_template)
        rcv_eat_template = spade.template.Template()
        rcv_eat_template.set_metadata("type", "eat")
        rcv_eat_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.rcv_eat(), rcv_eat_template)
        recv_broadcast_template = spade.template.Template()
        recv_broadcast_template.set_metadata("type", "broadcast")
        recv_broadcast_template.set_metadata("performative", "Inform")
        self.add_behaviour(self.recv_broadcast(), recv_broadcast_template)
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after setup: {self.__dict__}')
    
    class BackupBehaviour(spade.behaviour.PeriodicBehaviour):
        def __init__(self, start_at, period):
            super().__init__(start_at=start_at, period=period)
            self.http_client = httpx.AsyncClient(timeout=period)
        
        async def run(self):
            data = {
                "__timestamp__": int(datetime.datetime.timestamp(datetime.datetime.utcnow())),
                "jid": str(self.agent.jid),
                "type": "wolf",
                "floats": {
                    "msgRCount": self.agent.msgRCount,
                    "msgSCount": self.agent.msgSCount,
                    "connCount": self.agent.connCount,
                    "energy": self.agent.energy,
                    "x": self.agent.x,
                    "y": self.agent.y,
                    "heading": self.agent.heading,
                },
                "enums": {
                    "invert": self.agent.invert,
                    "dead": self.agent.dead,
                    "is_eating": self.agent.is_eating,
                },
                "connections": {
                    "connections": self.agent.connections,
                    "field": self.agent.field,
                    "sheep": self.agent.sheep,
                },
                "messages": {
                },
                "float_lists": {
                },
            }
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Sending backup data: {data}')
            try:
                await self.http_client.post(self.agent.backup_url, headers={"Content-Type": "application/json"}, data=orjson.dumps(data))
            except Exception as e:
                if self.agent.logger: self.agent.logger.error(f'[{self.agent.jid}] Backup error type: {e.__class__}, additional info: {e}')
    
    class setup(spade.behaviour.OneShotBehaviour):
        async def broadcast(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action broadcast')
            send = { "type": "broadcast", "performative": "Inform", "typ": 0.0, "state": 0.0, }
            
            # set
            send["typ"] = self.agent.limit_number(2)
            
            # if equal
            if self.agent.dead == "YES":
                
                # set
                send["state"] = self.agent.limit_number(0)
            
            # if equal
            if self.agent.dead == "NO":
                
                # set
                send["state"] = self.agent.limit_number(1)
            
            # send
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.connections}')
            for receiver in self.agent.connections:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            await self.broadcast()
    
    class move_around(spade.behaviour.PeriodicBehaviour):
        def wiggle(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action wiggle')
            
            # if equal
            if self.agent.dead == "NO":
                
                # float declaration
                wiglR = self.agent.limit_number(0)
                
                # float declaration
                wiglL = self.agent.limit_number(0)
                
                # uniform distribution
                wiglR = self.agent.limit_number(random.uniform(self.agent.limit_number(0), self.agent.limit_number(90)))
                
                # uniform distribution
                wiglL = self.agent.limit_number(random.uniform(self.agent.limit_number(0), self.agent.limit_number(90)))
                
                # float declaration
                change_flag = self.agent.limit_number(0)
                
                # if equal
                if self.agent.invert == "YES":
                    
                    # set
                    change_flag = self.agent.limit_number(1)
                    
                    # add
                    self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(wiglL))
                    
                    # subtract
                    self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(wiglR))
                    
                    # if less than
                    if self.agent.limit_number(self.agent.heading) < self.agent.limit_number(0):
                        
                        # set
                        self.agent.invert = "NO"
                        
                        # add
                        self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(180))
                    
                    # if greater than
                    if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(180):
                        
                        # set
                        self.agent.invert = "NO"
                        
                        # subtract
                        self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(180))
                
                # if equal
                if self.agent.invert == "NO":
                    
                    # if equal
                    if self.agent.limit_number(change_flag) == self.agent.limit_number(0):
                        
                        # add
                        self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(wiglL))
                        
                        # subtract
                        self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(wiglR))
                        
                        # if less than
                        if self.agent.limit_number(self.agent.heading) < self.agent.limit_number(0):
                            
                            # set
                            self.agent.invert = "YES"
                            
                            # add
                            self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(180))
                        
                        # if greater than
                        if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(180):
                            
                            # set
                            self.agent.invert = "YES"
                            
                            # subtract
                            self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(180))
        
        def move(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action move')
            
            # if equal
            if self.agent.dead == "NO":
                
                # float declaration
                a = self.agent.limit_number(0)
                
                # float declaration
                b = self.agent.limit_number(0)
                
                # sin
                a = self.agent.limit_number(numpy.sin(numpy.deg2rad(self.agent.limit_number(self.agent.heading))))
                
                # sin
                b = self.agent.limit_number(numpy.sin(numpy.deg2rad(self.agent.limit_number(self.agent.heading))))
                
                # float declaration
                newx = self.agent.limit_number(self.agent.x)
                
                # float declaration
                newy = self.agent.limit_number(self.agent.y)
                
                # if greater than
                if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(90):
                    
                    # subtract
                    newx = self.agent.limit_number(newx - self.agent.limit_number(a))
                
                # if greater than or equal
                if self.agent.limit_number(self.agent.heading) <= self.agent.limit_number(90):
                    
                    # add
                    newx = self.agent.limit_number(newx + self.agent.limit_number(a))
                
                # if equal
                if self.agent.invert == "YES":
                    
                    # subtract
                    newy = self.agent.limit_number(newy - self.agent.limit_number(b))
                
                # if equal
                if self.agent.invert == "NO":
                    
                    # add
                    newy = self.agent.limit_number(newy + self.agent.limit_number(b))
                
                # float declaration
                bump = self.agent.limit_number(0)
                
                # if less than
                if self.agent.limit_number(newx) < self.agent.limit_number(0):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if less than
                if self.agent.limit_number(newy) < self.agent.limit_number(0):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if greater than
                if self.agent.limit_number(newx) > self.agent.limit_number(99):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if greater than
                if self.agent.limit_number(newy) > self.agent.limit_number(99):
                    
                    # set
                    bump = self.agent.limit_number(1)
                
                # if equal
                if self.agent.limit_number(bump) == self.agent.limit_number(0):
                    
                    # set
                    self.agent.x = self.agent.limit_number(newx)
                    
                    # set
                    self.agent.y = self.agent.limit_number(newy)
                    
                    # subtract
                    self.agent.energy = self.agent.limit_number(self.agent.energy - self.agent.limit_number(1))
                
                # if equal
                if self.agent.limit_number(bump) == self.agent.limit_number(1):
                    
                    # add
                    self.agent.heading = self.agent.limit_number(self.agent.heading + self.agent.limit_number(90))
                    
                    # if greater than
                    if self.agent.limit_number(self.agent.heading) > self.agent.limit_number(180):
                        
                        # subtract
                        self.agent.heading = self.agent.limit_number(self.agent.heading - self.agent.limit_number(180))
                        
                        # float declaration
                        flag = self.agent.limit_number(0)
                        
                        # if equal
                        if self.agent.invert == "YES":
                            
                            # set
                            flag = self.agent.limit_number(1)
                            
                            # set
                            self.agent.invert = "NO"
                        
                        # if equal
                        if self.agent.invert == "NO":
                            
                            # if equal
                            if self.agent.limit_number(flag) == self.agent.limit_number(0):
                                
                                # set
                                self.agent.invert = "YES"
        
        def check_if_dead(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action check_if_dead')
            
            # if greater than or equal
            if self.agent.limit_number(self.agent.energy) <= self.agent.limit_number(0):
                
                # set
                self.agent.dead = "YES"
        
        async def eat_sheep(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action eat_sheep')
            send = { "type": "eat", "performative": "Request", "x": 0.0, "y": 0.0, }
            
            # set
            send["x"] = self.agent.limit_number(self.agent.x)
            
            # set
            send["y"] = self.agent.limit_number(self.agent.y)
            
            # send
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.sheep}')
            for receiver in self.agent.sheep:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
            
            # set
            self.agent.is_eating = "YES"
        
        async def run(self):
            self.wiggle()
            self.move()
            self.check_if_dead()
            await self.eat_sheep()
    
    class recv_spawn(spade.behaviour.CyclicBehaviour):
        def spawn(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action spawn')
            
            # set
            self.agent.dead = "NO"
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.spawn(rcv)
    
    class rcv_eat(spade.behaviour.CyclicBehaviour):
        async def eat_sheep(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action eat_sheep')
            send = { "type": "kill", "performative": "Inform", }
            
            # if equal
            if self.agent.is_eating == "YES":
                
                # set
                self.agent.is_eating = "NO"
                
                # send
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {"RCV.jid"}')
                await self.send(self.agent.get_spade_message("RCV.jid", send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                await self.eat_sheep(rcv)
    
    class recv_broadcast(spade.behaviour.CyclicBehaviour):
        def save_data(self, rcv):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action save_data')
            
            # if equal
            if self.agent.limit_number(rcv["typ"]) == self.agent.limit_number(0):
                
                # add element
                self.agent.field.append("RCV.jid")
            
            # if equal
            if self.agent.limit_number(rcv["typ"]) == self.agent.limit_number(1):
                
                # add element
                self.agent.sheep.append("RCV.jid")
        
        async def run(self):
            rcv = await self.receive(timeout=100000)
            if rcv:
                rcv = self.agent.get_json_from_spade_message(rcv)
                self.agent.msgRCount = self.agent.limit_number(self.agent.msgRCount + 1)
                if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Received message: {rcv}')
                self.save_data(rcv)
    

import random
import uuid
import numpy


def generate_graph_structure(domain):
    _num_grass = 1
    _num_sheep = 95
    _num_wolf = 5
    num_agents = _num_grass + _num_sheep + _num_wolf
    random_id = str(uuid.uuid4())[:5]
    jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]
    agents = []
    next_agent_idx = 0
    for _ in range(_num_grass):
        num_connections = 100
        num_connections = max(min(num_connections, len(jids) - 1), 0)
        jid = jids[next_agent_idx]
        agents.append({
            "jid": jid,
            "type": "grass",
            "connections": random.sample([other_jid for other_jid in jids if other_jid != jid], num_connections),
        })
        next_agent_idx += 1
    for _ in range(_num_sheep):
        num_connections = 100
        num_connections = max(min(num_connections, len(jids) - 1), 0)
        jid = jids[next_agent_idx]
        agents.append({
            "jid": jid,
            "type": "sheep",
            "connections": random.sample([other_jid for other_jid in jids if other_jid != jid], num_connections),
        })
        next_agent_idx += 1
    for _ in range(_num_wolf):
        num_connections = 100
        num_connections = max(min(num_connections, len(jids) - 1), 0)
        jid = jids[next_agent_idx]
        agents.append({
            "jid": jid,
            "type": "wolf",
            "connections": random.sample([other_jid for other_jid in jids if other_jid != jid], num_connections),
        })
        next_agent_idx += 1
    return agents



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.graph import (
    AgentConstantAmount,
    AgentPercentAmount,
    ConnectionConstantAmount,
    ConnectionDistExpAmount,
    ConnectionDistNormalAmount,
    ConnectionDistUniformAmount,
    StatisticalAgent,
    StatisticalGraph,
)
from aasm.utils.validation import is_float, is_int

if TYPE_CHECKING:
    from parsing.state import State

    from aasm.intermediate.graph import AgentAmount, ConnectionAmount


def op_DEFG(state: State, agent_name: str, amount: str, args: List[str]) -> None:
    state.require(
        state.in_graph,
        "Cannot define agent graph amount outside graph scope.",
        "Try defining new graphs using GRAPH.",
    )
    state.require(
        isinstance(state.last_graph, StatisticalGraph),
        "DEFG can be used with statistical graphs.",
        "Define statistical graphs with GRAPH statistical.",
    )
    state.require(state.agent_exists(agent_name), f"Agent {agent_name} is not defined.")

    agent_amount: AgentAmount | None = None
    if amount.endswith("%"):
        percent_value = amount[:-1]
        state.require(is_float(percent_value), f"{amount} is not a valid float.")
        state.require(
            float(percent_value) >= 0 and float(percent_value) <= 100,
            f"{amount} is not a valid agent amount.",
            "Amount must be non-negative.",
        )

        agent_amount = AgentPercentAmount(percent_value)

    else:
        state.require(is_int(amount), f"{amount} is not a valid integer.")
        state.require(
            int(amount) >= 0,
            f"{amount} is not a valid agent amount.",
            "Amount must be non-negative.",
        )

        agent_amount = AgentConstantAmount(amount)

    connection_amount: ConnectionAmount | None = None
    match args:
        case [value]:
            state.require(is_int(value), f"{value} is not a valid integer.")
            state.require(
                int(value) >= 0,
                f"{value} is not a valid connection amount.",
                "Amount must be non-negative.",
            )

            connection_amount = ConnectionConstantAmount(value)

        case ["dist_normal", mean, std_dev]:
            state.require(is_float(mean), f"{mean} is not a valid float.")
            state.require(is_float(std_dev), f"{std_dev} is not a valid float.")
            state.require(
                float(std_dev) >= 0,
                f"{std_dev} is not a valid standard deviation parameter.",
                "Standard deviation must be non-negative.",
            )

            connection_amount = ConnectionDistNormalAmount(mean, std_dev)

        case ["dist_exp", lambda_]:
            state.require(is_float(lambda_), f"{lambda_} is not a valid float.")
            state.require(
                float(lambda_) > 0,
                f"{lambda_} is not a valid lambda parameter.",
                "Lambda must be positive.",
            )

            connection_amount = ConnectionDistExpAmount(lambda_)

        case ["dist_uniform", a, b]:
            state.require(is_float(a), f"{a} is not a valid float.")
            state.require(is_float(b), f"{b} is not a valid float.")

            connection_amount = ConnectionDistUniformAmount(a, b)

        case _:
            state.panic(f"Incorrect operation: DEFG {agent_name} {amount} {args}")

    state.last_graph.add_agent(
        StatisticalAgent(agent_name, agent_amount, connection_amount)
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import ListWrite

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_LW(state: State, arg1: str, arg2: str, arg3: str) -> None:
    state.require(
        state.in_action, "Not inside any action.", "LW can be used inside actions."
    )
    list_ = Argument(state, arg1)
    idx = Argument(state, arg2)
    value = Argument(state, arg3)
    state.require(
        list_.list_write_context(idx, value),
        "Mismatched types in the list write context.",
        f"ARG1 {list_.explain()}, ARG2 {idx.explain()}, ARG3 {value.explain()}",
    )

    state.last_action.add_instruction(ListWrite(list_, idx, value))
